/*
    Copyright (C) 2004 Jesse Chappell
    Copyright (C) 2003 Jack O'Quin
    Copyright (C) 2001 Paul Davis
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <errno.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>

#include <iostream>

#include <string>

#include "audio_driver.hpp"
#include "jack_audio_driver.hpp"
#include "engine.hpp"

#include "version.h"


extern "C" {
  static void atexit_cleanup(void);
  static void* watchdog_thread(void* arg); 
  static void dummy_signal_handler(int signal);
}

static void setup_signals();

using namespace std;
using namespace AC3JACK;

struct OptionInfo
{
	OptionInfo() :
		quiet(false), jack_name(""),
		show_usage(false), show_version(false) {} 
		
	bool   quiet;
	string jack_name;
	string jack_server_name;
	int    show_usage;
	bool   show_version;
};

static Engine * engine = 0;

void usage(char *argv0)
{
	EngineParams dummyparams;

	fprintf(stderr, "ac3jack %s\nCopyright 2004 Jesse Chappell\nac3jack comes with ABSOLUTELY NO WARRANTY\n", ac3jack_version);
	fprintf(stderr, "This is free software, and you are welcome to redistribute it\n");
	fprintf(stderr, "under certain conditions; see the file COPYING for details\n\n");

	//fprintf(stderr, "Usage: %s [options...] [portname ...]\n", argv0);
	fprintf(stderr, "Usage: %s [options...]\n", argv0);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -o <str> , --output=<str>    output filename for AC3 file.  If 'stdout', then it is sent to standard output.\n");
	fprintf(stderr, "  -c <num> , --channels=<num>  number of AC-3 output channels (1-5). default is 5\n");
	fprintf(stderr, "  -l       , --lfe             enable LFE channel\n");
	fprintf(stderr, "  -b <num> , --bitrate=<num>   bitrate of AC-3 stream (kbits/s). default is 448\n");
	fprintf(stderr, "  -d <num> , --duration=<num>  duration of output stream (secs). default is 0 (forever)\n");
	fprintf(stderr, "  -t , --jack-transport        use jack transport state (rolling/stopped) to control stream output\n");
	fprintf(stderr, "  -R , --raw-jack-out          will write raw deinterleaved ac3 out two jack ports.  careful!\n");
	fprintf(stderr, "  -S , --spdif-out             will open a local audio device SPDIF for live AC3 output\n");
	fprintf(stderr, "  -C <str> , --spdif-out-card  optionally specify the card device name to use with -S option (ALSA only)\n");
	fprintf(stderr, "  -j <str> , --jack-name=<str> jack client name, default is ac3jack\n");
	//fprintf(stderr, "  -B <num> , --bufsize=<num>   internal buffer size (samples), if overrun problems occur try a large value\n");
	fprintf(stderr, "  -P <num> , --osc-port=<num>  OSC server port to use (default is %d)\n", dummyparams.osc_port);
	fprintf(stderr, "  -q , --quiet                 do not output status to stderr\n");
	fprintf(stderr, "  -h , --help                  this usage output\n");
	fprintf(stderr, "  -V , --version               show version only\n");
	//fprintf(stderr, "\nIf portname arguments are added, they are initially connected in channel order, if possible.\n");
}

static void parse_options (int argc, char **argv, OptionInfo & option_info, EngineParams & engine_params)
{
	int c;
	int longopt_index = 0;
	extern int optind, opterr;

	char *optstring = "c:d:o:b:B:hqVtRj:SC:P:l";
	struct option long_options[] = {
		{ "help", 0, 0, 'h' },
		{ "quiet", 0, 0, 'q' },
		{ "duration", 1, 0, 'd' },
		{ "channels", 1, 0, 'c' },
		{ "lfe", 0, 0, 'l' },
		{ "output", 1, 0, 'o' },
		{ "bitrate", 1, 0, 'b' },
		{ "bufsize", 1, 0, 'B' },
		{ "osc-port", 1, 0, 'P' },
		{ "jack-name", 1, 0, 'j' },
		{ "jack-transport", 0, 0, 't' },
		{ "raw-jack-out", 0, 0, 'R' },
		{ "spdif-out", 0, 0, 'S' },
		{ "spdif-out-card", 1, 0, 'C' },
		{ "version", 0, 0, 'V' },
		{ 0, 0, 0, 0 }
	};

	opterr = 0;

	while ((c = getopt_long (argc, argv, optstring, long_options, &longopt_index)) >= 0) {
		if (c >= 255) {
			break;
		}
		
		switch (c) {
		case 1:
			/* getopt signals end of '-' options */
			break;

		case 'h':
			option_info.show_usage++;
			break;
		case 'V':
			option_info.show_version++;
			break;
		case 'q':
			option_info.quiet = true;
			break;
		case 'j':
			option_info.jack_name = optarg;
			break;
		case 't':
		        engine_params.use_transport = true;
			break;
		case 'd':
			engine_params.duration = atoi (optarg);
			break;
		case 'c':
		        engine_params.channels = atoi(optarg);
			break;
		case 'P':
		        engine_params.osc_port = atoi(optarg);
			break;
		case 'C':
		        engine_params.spdif_dev = optarg;
			break;
		case 'b':
			engine_params.bitrate = atoi (optarg);
			break;
		case 'R':
			engine_params.raw_jackout = true;
			break;
		case 'S':
			engine_params.spdif_out = true;
			break;
		case 'o':
			engine_params.outfile_name = optarg;
			engine_params.file_output = true;
			break;
		case 'l':
			engine_params.lfe = true;
			break;
			
		default:
			fprintf (stderr, "argument error: %d\n", c);
			option_info.show_usage++;
			break;
		}

		if (option_info.show_usage > 0) {
			break;
		}
	}

	// argc - optind, &argv[optind]
	for (int i=0; i < argc-optind; ++i) {
		engine_params.portconns.push_back(argv[optind + i]);
	}
}

int
main (int argc, char *argv[])

{
	OptionInfo option_info;
	EngineParams engine_params;
	
	parse_options (argc, argv, option_info, engine_params);

	if (engine_params.channels <= 0) {
		engine_params.channels = 1;
	}

	if (option_info.show_usage ) {
		usage(argv[0]);
		exit (1);
	}
	
	if (!option_info.quiet || option_info.show_version) {
		fprintf(stderr, "ac3jack %s  -- Copyright 2007 Jesse Chappell\n", ac3jack_version);
	}

	if (option_info.show_version) {
		exit(0);
	}

	setup_signals();
	
	AudioDriver * driver = new JackAudioDriver(option_info.jack_name, option_info.jack_server_name);

	engine = new Engine();

	engine->set_params (engine_params);
	
	if (!engine->initialize(driver)) {
		cerr << "cannot initialize ac3jack\n";
		exit (1);
	}

	if (!driver->activate()) {
		exit(1);
	}


	// go into engine's non-rt event loop
	// this returns when we quit or signal handler causes it to
	engine->mainloop();

	driver->deactivate();
	usleep(200000);
	driver->set_engine(0);
	
	delete engine;
	engine = 0;
	
	delete driver;

	cerr << "exiting" << endl;

	return 0;
}


static void atexit_cleanup(void)
{
  struct sigaction blockaction;
  sigset_t signalset;

  blockaction.sa_handler = SIG_DFL;
  sigemptyset(&blockaction.sa_mask);
  blockaction.sa_flags = 0;

  /* set back to default the following signals */
  sigaction(SIGTERM, &blockaction, 0);
  sigaction(SIGINT, &blockaction, 0);
  sigaction(SIGHUP, &blockaction, 0);
  sigaction(SIGPIPE, &blockaction, 0);
  sigaction(SIGABRT, &blockaction, 0);

  /* also must mask the signals for the watchdog handler to get signalled on OSX */
  sigemptyset(&signalset);
  
  if (sigprocmask(SIG_SETMASK, &signalset, NULL) != 0) {
	  cerr << "error setting mask" << endl;
  }
  
  cerr << "cleanup atexit!" << endl;
}

static void* watchdog_thread(void* arg)
{
  sigset_t signalset;
  int signalno;
  int exiting = 0;

  /* register cleanup routine */
  atexit(&atexit_cleanup);

  // unmask these
  sigemptyset(&signalset);

  if (pthread_sigmask(SIG_SETMASK, &signalset, NULL) != 0) {
	  cerr << "couldn't set mask" << endl;
  }

  while (!exiting)
  {
	  sigemptyset(&signalset);
	  
	  /* handle the following signals explicitly */
	  sigaddset(&signalset, SIGTERM);
	  sigaddset(&signalset, SIGINT);
	  sigaddset(&signalset, SIGHUP);
	  sigaddset(&signalset, SIGPIPE);

	  /* block until a signal received */
	  sigwait(&signalset, &signalno);
	  
	  fprintf (stderr, "recieved signal %d\n", signalno);
	  
	   exiting = 1;
  }

  if (engine) {
	  engine->quit();
  }

  return(0);
}



/**
 * Sets up a signal mask with sigaction() that blocks 
 * all common signals, and then launces an watchdog
 * thread that waits on the blocked signals using
 * sigwait().
 */
static void setup_signals()
{
  pthread_t watchdog;
  pthread_attr_t pattr;
  int res;
  sigset_t signalset;

  /* man pthread_sigmask:
   *  "...signal actions and signal handlers, as set with
   *   sigaction(2), are shared between all threads"
   */

  struct sigaction blockaction;
#ifdef __APPLE__
  blockaction.sa_handler = dummy_signal_handler;
#else
  blockaction.sa_handler = SIG_IGN;  
#endif
  sigemptyset(&blockaction.sa_mask);
  blockaction.sa_flags = 0;

  /* ignore the following signals */

  sigaction(SIGTERM, &blockaction, 0);
  sigaction(SIGINT, &blockaction, 0);
  sigaction(SIGHUP, &blockaction, 0);
  sigaction(SIGPIPE, &blockaction, 0);
  sigaction(SIGABRT, &blockaction, 0);

//#ifdef __APPLE__
  /* also must mask the signals for the watchdog handler to get signalled on OSX */
  sigemptyset(&signalset);
  sigaddset(&signalset, SIGTERM);
  sigaddset(&signalset, SIGINT);
  sigaddset(&signalset, SIGHUP);
  sigaddset(&signalset, SIGPIPE);

  if (sigprocmask(SIG_SETMASK, &signalset, NULL) != 0) {
	  cerr << "error setting mask" << endl;
  }
//#endif
  
  pthread_attr_init(&pattr);
  pthread_attr_setdetachstate(&pattr, PTHREAD_CREATE_DETACHED);

  
  res = pthread_create(&watchdog, 
			   &pattr, 
			   watchdog_thread, 
			   NULL);
  if (res != 0) {
     fprintf(stderr, "ac3jack: Warning! Unable to create watchdog thread.\n");
  }

}

static void dummy_signal_handler(int signal)
{
  cerr << " WARNING! dummy_signal_handler entered, this should _NOT_ happen!";
  cerr << " pid=" << getpid() << " signal: " << signal << " thread: " << (int) pthread_self() << endl;
}
