/*
** Copyright (C) 2006 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#include <iostream>
#include "boost/date_time/posix_time/posix_time.hpp"
#include "boost/thread/thread.hpp"
#include "boost/thread/xtime.hpp"

#include <cstdio>
#include <cerrno>
#include <assert.h>
#include <alloca.h>
#include <cstring>
#include <unistd.h>

#include "alsa_spdif_writer.hpp"


using namespace AC3JACK;
using namespace std;
using namespace boost;

AlsaSpdifWriter::AlsaSpdifWriter( unsigned int samplerate, const std::string & cardstr, size_t bufsize)
	: SpdifWriter(), _cardstr(cardstr), _samplerate(samplerate), _allow_pcm(false)
{
	_pcm = 0;
	_buffer = new RingBuffer<unsigned char>(bufsize);

	probe_devices();
		
}

AlsaSpdifWriter::~AlsaSpdifWriter()
{
	// ensure thread is done
	if (_active) {
		_active = false;
		
		{
			boost::try_mutex::scoped_lock lk(_butler_mutex);
			_more_incoming.notify_one();
		}

		cerr << "AS waiting for butler thread to end in alsa spdif writer" << endl;
		_butler_thread->join();
		cerr << "joined butler" << endl;
	}

	delete _buffer;

	close_device();
	
}

void AlsaSpdifWriter::start()
{
	open_device();

	SpdifWriter::start();
}

bool AlsaSpdifWriter::open_device()
{
	if (_pcm) {
		close_device();
	}
	
	boost::try_mutex::scoped_lock lk(_butler_mutex);

	snd_pcm_hw_params_t *params;
	unsigned int rate, buffer_time, period_time, tmp;
	snd_pcm_format_t format = SND_PCM_FORMAT_S16;
	int err, step;
	char pcm_name[200];
	char errstr[200];
	errstr[0] = '\0';
	
	snd_pcm_hw_params_alloca(&params);

	_ok = false;

	if (_cardstr.empty()) {
		snprintf(pcm_name, sizeof(pcm_name), "plug:iec958:{AES0 0x2 AES1 0x82 AES2 0x0 AES3 0x2 }");
	}
	else {
		//snprintf(pcm_name, sizeof(pcm_name), "plug:iec958:{AES0 0x2 AES1 0x82 AES2 0x0 AES3 0x2 CARD %s }", _cardstr.c_str());
		snprintf(pcm_name, sizeof(pcm_name), "%s", _cardstr.c_str());
	}
	
	/*
	 * Open the device driver
	 */
	fprintf(stdout, "Using PCM device '%s'\n", pcm_name);
	
	if ((err = snd_pcm_open(&_pcm, pcm_name, SND_PCM_STREAM_PLAYBACK,  SND_PCM_NONBLOCK)) < 0) {
		snprintf(errstr, sizeof(errstr), "Error opening '%s': %s\n", pcm_name, snd_strerror(err));
		_status = errstr;
		return err;
	}
	cerr << "opened device: " << pcm_name <<  endl;
	err = snd_pcm_hw_params_any(_pcm, params);
	if (err < 0) {
		snprintf(errstr, sizeof(errstr), "Broken configuration for this PCM: no configurations available\n");
		goto __close;
	}
	/* set interleaved access */
	err = snd_pcm_hw_params_set_access(_pcm, params,
					   SND_PCM_ACCESS_RW_INTERLEAVED);
	if (err < 0) {
		snprintf(errstr, sizeof(errstr), "Error: Access type not available\n");
		goto __close;
	}
	err = snd_pcm_hw_params_set_format(_pcm, params, format);
	if (err < 0) {
		snprintf(errstr, sizeof(errstr), "Error: Sample format %d not available\n", format);
		goto __close;
	}
	err = snd_pcm_hw_params_set_channels(_pcm, params, 2);
	if (err < 0) {
		snprintf(errstr, sizeof(errstr), "Error: Channels count %d not available\n", 2);
		goto __close;
	}
	rate = _samplerate;
	err = snd_pcm_hw_params_set_rate_near(_pcm, params, &rate, 0);
	if (err < 0) {
		snprintf(errstr, sizeof(errstr), "Error: Rate %d not available\n", rate);
		goto __close;
	}
	cerr << "using rate:" << rate << endl;
	
	buffer_time = 500000;
	err = snd_pcm_hw_params_set_buffer_time_near(_pcm, params, &buffer_time, 0);
	if (err < 0) {
		snprintf(errstr, sizeof(errstr), "Error: Buffer time %d not available\n", buffer_time);
		goto __close;
	}
	cerr << "using buffer time:" << buffer_time << endl;

	step = 2;
	period_time = 10000 * 2;
	do {
		period_time /= 2;
		tmp = period_time;
		err = snd_pcm_hw_params_set_period_time_near(_pcm, params, &tmp, 0);
		if (err < 0) {
			snprintf(errstr, sizeof(errstr), "Error: Period time %d not available\n", tmp);
			goto __close;
		}
		if (tmp == period_time) {
			period_time /= 3;
			tmp = period_time;
			err = snd_pcm_hw_params_set_period_time_near(_pcm, params, &tmp, 0);
			if (tmp == period_time)
				period_time = 10000 * 2;
		}
	} while (buffer_time == period_time && period_time > 10000);
	cerr << "using period time:" << period_time << endl;

	if (buffer_time == period_time) {
		snprintf(errstr, sizeof(errstr), "Error: Buffer time and period time match, could not use\n");
		goto __close;
	}
	if ((err = snd_pcm_hw_params(_pcm, params)) < 0) {
		snprintf(errstr, sizeof(errstr), "Error: PCM hw_params failed: %s\n", snd_strerror(err));
		goto __close;
	}

	_ok = true;

	_status = "Active";
	
	return _ok;
	
      __close:

	_status = errstr;
	
      	snd_pcm_close(_pcm);
      	_pcm = NULL;
	_ok = false;

	return _ok;
}

void AlsaSpdifWriter::close_device()
{
	boost::try_mutex::scoped_lock lk(_butler_mutex);

	if (_pcm) {
		snd_pcm_close(_pcm);
		_pcm = NULL;
		_ok = false;
	}
}

void AlsaSpdifWriter::probe_devices()
{
	// Enumerate the ALSA cards and PCM hardware devices...
	snd_ctl_t *handle;
	snd_ctl_card_info_t *info;
	snd_pcm_info_t *pcminfo;
	snd_ctl_card_info_alloca(&info);
	snd_pcm_info_alloca(&pcminfo);
	int iCard = -1;
	char cardstr[100];
	char tmpstr[100];
	char devstr[100];		

	_devlist.clear();

	snprintf(cardstr, sizeof(cardstr), "plug:iec958:{AES0 0x2 AES1 0x82 AES2 0x0 AES3 0x2}");
	
	if (snd_ctl_open(&handle, cardstr, 0) >= 0
	    && snd_ctl_card_info(handle, info) >= 0)
	{
		snprintf(tmpstr, sizeof(tmpstr), "%s (%s)", snd_ctl_card_info_get_name(info), "AC3" );
		_devlist.push_back(DeviceInfo(string(tmpstr), string(cardstr)));
		snd_ctl_close(handle);
	}
	
	while (snd_card_next(&iCard) >= 0 && iCard >= 0)
	{
		bool found_real = false;
		snprintf(cardstr, sizeof(cardstr), "plug:iec958:{AES0 0x2 AES1 0x82 AES2 0x0 AES3 0x2 CARD %d }", iCard);

		if (snd_ctl_open(&handle, cardstr, 0) >= 0
		    && snd_ctl_card_info(handle, info) >= 0)
		{
			snprintf(tmpstr, sizeof(tmpstr), "%s (%s)", snd_ctl_card_info_get_name(info), "AC3" );
			_devlist.push_back(DeviceInfo(string(tmpstr), string(cardstr)));
			found_real = true;
			snd_ctl_close(handle);
		}
		
		if (_allow_pcm && !found_real)
		{
			snprintf(cardstr, sizeof(cardstr), "hw:%d", iCard);

			if (snd_ctl_open(&handle, cardstr, 0) >= 0
			    && snd_ctl_card_info(handle, info) >= 0)
			{
				int iDevice = -1;

				//snprintf(tmpstr, sizeof(tmpstr), "%s (%s)", snd_ctl_card_info_get_name(info), cardstr );
				//	_devlist.push_back(DeviceInfo(string(tmpstr), string(cardstr)));
				
				//unsigned int channels_max ;
				//err = snd_pcm_hw_params_get_channels_max (hw_params, &channels_max);
 
				while (snd_ctl_pcm_next_device(handle, &iDevice) >= 0 
				       && iDevice >= 0)
				{
					snd_pcm_info_set_device(pcminfo, iDevice);
					snd_pcm_info_set_subdevice(pcminfo, 0);
					snd_pcm_info_set_stream(pcminfo, SND_PCM_STREAM_PLAYBACK);
					if (snd_ctl_pcm_info(handle, pcminfo) >= 0)
					{
						snprintf(devstr, sizeof(devstr), "%s,%d", cardstr, iDevice);
						snprintf(tmpstr, sizeof(tmpstr), "%s (%s)", snd_ctl_card_info_get_name(info), devstr );

						_devlist.push_back(DeviceInfo(string(tmpstr), string(devstr)));
					}
				}
				
				snd_ctl_close(handle);
				//++iCards;
			}
		}
	}
}


size_t AlsaSpdifWriter::write(unsigned char * data, size_t length)
{
	size_t wrote;

	wrote = _buffer->write (data, length);

	if (length != wrote) {
		// cerr << "ONLY wrote " << wrote << "  of " << length << endl;
	}
	
	// non blocking try lock
	boost::try_mutex::scoped_try_lock lk(_butler_mutex);
	if (lk) {
		_more_incoming.notify_one();
	}
					

	return wrote;
}

// write thread routines follow

void AlsaSpdifWriter::run_work_thread()
{
	const size_t MIN_BLOCKSIZE = 0;
	
	// this is released when we are waiting for more to write
	boost::try_mutex::scoped_lock lk(_butler_mutex);

	cerr << "started alsa spdif writer work thread" << endl;
	
	while (_active)
	{
                // sleep till we are needed
                wait_for_more(lk);

                if (!_active) {
                        break;
                }
		else if (!_pcm) {
			continue;
		}
		
		write_to_spdif (MIN_BLOCKSIZE);
		
	}

	// time to stop, make sure to write out any remaining stuff

	write_to_spdif (0);

	cerr << "finished alsa spdif writer work thread" << endl;	
}

void AlsaSpdifWriter::write_to_spdif(size_t minblocksize)
{
	RingBuffer<unsigned char>::rw_vector readvec;

	if (_pcm)
	{
		while (_buffer->read_space() > minblocksize) {
				
			_buffer->get_read_vector(&readvec);

			// cerr << "got data to output: " << readvec.len[0] + readvec.len[1] << endl;
			
			// stereo 16bit
			output_play ((short *) readvec.buf[0], readvec.len[0] / 4);
			
			if (readvec.len[1] > 0) {
				output_play ((short *) readvec.buf[1], readvec.len[1] / 4);
			}

			_buffer->increment_read_ptr (readvec.len[0] + readvec.len[1]);
		}
	}
}

/*
 * play the sample to the already opened file descriptor
 */
int AlsaSpdifWriter::output_play(short * output_samples, nframes_t num_frames)
{
	snd_pcm_sframes_t res = 0;

	do {
		if (res == -EPIPE) {		/* underrun */
			res = snd_pcm_prepare(_pcm);
			// cerr << "UNderreun " << endl;
		}
		else if (res == -ESTRPIPE) {	/* suspend */
			while ((res = snd_pcm_resume(_pcm)) == -EAGAIN)
				sleep(1);
			if (res < 0)
				res = snd_pcm_prepare(_pcm);
		}
		res = res < 0 ? res : snd_pcm_writei(_pcm, (void *)output_samples, num_frames);
		if (res > 0) {
			output_samples += 2 * res;
			num_frames -= res;
		}
	} while (res == -EPIPE || num_frames > 0);
	if (res < 0)
		fprintf(stderr, "writei returned error: %s\n", snd_strerror(res));
	return res < 0 ? (int)res : 0;
}


void AlsaSpdifWriter::wait_for_more(try_mutex::scoped_lock & lk)
{
	// we just need to sleep until more data shows up
	// timeout just in case
	xtime sleeptime;

        while (_buffer->read_space() == 0 && _active) {
                xtime_get(&sleeptime, TIME_UTC);
                sleeptime.nsec += 100000000;
                if (sleeptime.nsec >= 1000000000) {
                        sleeptime.sec += 1;
                        sleeptime.nsec -= 1000000000;
                }
                _more_incoming.timed_wait (lk, sleeptime);
        }

}

int AlsaSpdifWriter::get_devices(DeviceInfoList & retlist)
{
	probe_devices();

	retlist.insert(retlist.begin(), _devlist.begin(), _devlist.end());
	return 0;
}


void AlsaSpdifWriter::set_allow_pcm (bool val)
{
	_allow_pcm = val;
}

void AlsaSpdifWriter::set_device (std::string devstr)
{
	if (_cardstr != devstr) {
		cerr << "Set device to " << devstr << endl;
		_cardstr = devstr;
		// TODO
	}
}
