/*
** Copyright (C) 2007 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cerrno>
#include <algorithm>

#include <sys/poll.h>
#include <unistd.h>
#include <fcntl.h>

#include "control_osc.hpp"
#include "engine.hpp"
#include "ringbuffer.hpp"
#include "version.h"

#include <lo/lo.h>
#include <sigc++/sigc++.h>
#include <string.h>
using namespace SigC;

using namespace AC3JACK;
using namespace std;

static void error_callback(int num, const char *m, const char *path)
{
#ifdef DEBUG
	fprintf(stderr, "liblo server error %d in path %s: %s\n", num, path, m);
#endif
}



ControlOSC::ControlOSC(Engine * eng, unsigned int port)
	: _engine(eng), _port(port)
{
	char tmpstr[255];

	_ok = false;
	_shutdown = false;
	_osc_server = 0;
	_osc_unix_server = 0;
	_osc_thread = 0;
	_max_instance = 0;
	
	for (int j=0; j < 20; ++j) {
		snprintf(tmpstr, sizeof(tmpstr), "%d", _port);

		if ((_osc_server = lo_server_new (tmpstr, error_callback))) {
			break;
		}
#ifdef DEBUG		
		cerr << "can't get osc at port: " << _port << endl;
#endif
		_port++;
		continue;
	}

	/*** APPEARS sluggish for now
	     
	// attempt to create unix socket server too
	//snprintf(tmpstr, sizeof(tmpstr), "/tmp/sooperlooper_%d", getpid());
	snprintf(tmpstr, sizeof(tmpstr), "/tmp/sooperlooper_XXXXXX");
	int fd = mkstemp(tmpstr);
	if (fd >=0) {
		unlink(tmpstr);
		close(fd);
		_osc_unix_server = lo_server_new (tmpstr, error_callback);
		if (_osc_unix_server) {
			_osc_unix_socket_path = tmpstr;
		}
	}

	*/

	register_callbacks();

	// lo_server_thread_add_method(_sthread, NULL, NULL, ControlOSC::_dummy_handler, this);


}

void ControlOSC::start()
{
	if (!init_osc_thread()) {
		return;
	}
	_ok = true;
}

ControlOSC::~ControlOSC()
{
	if (!_osc_unix_socket_path.empty()) {
		// unlink it
		unlink(_osc_unix_socket_path.c_str());
	}

	// stop server thread
	terminate_osc_thread();
}

void
ControlOSC::register_callbacks()
{
	lo_server srvs[2];
	lo_server serv;
	char  path[255];
	char  types[10];
	
	srvs[0] = _osc_server;
	srvs[1] = _osc_unix_server;
	
	for (size_t i=0; i < 2; ++i) {
		if (!srvs[i]) continue;
		serv = srvs[i];


		/* add method that will match the path /quit with no args */
		lo_server_add_method(serv, "/quit", "", ControlOSC::_quit_handler, this);

		// add ping handler:   s:retpath [ s:returl ]
		lo_server_add_method(serv, "/ping", "ss", ControlOSC::_ping_handler, this);
		lo_server_add_method(serv, "/ping", "s", ControlOSC::_ping_handler, this);
		lo_server_add_method(serv, "/ping", "", ControlOSC::_ping_handler, this);

		// un/register param change handler:  s:retpath [ s:returl ]  
		lo_server_add_method(serv, "/register", "ss", ControlOSC::_register_update_handler, this);
		lo_server_add_method(serv, "/register", "s", ControlOSC::_register_update_handler, this);
		lo_server_add_method(serv, "/unregister", "ss", ControlOSC::_unregister_update_handler, this);
		lo_server_add_method(serv, "/unregister", "s", ControlOSC::_unregister_update_handler, this);

		// get all params
		lo_server_add_method(serv, "/params_all", "ss", ControlOSC::_param_get_all_handler, this);		
		lo_server_add_method(serv, "/params_all", "s", ControlOSC::_param_get_all_handler, this);

		// specific parameters

		const ParamMap & param_map = _engine->get_parameter_map();

		for (ParamMap::const_iterator iter = param_map.begin();
		     iter != param_map.end(); ++iter)
		{
			// get parameter s:retpath [s:returl]
			// set parameter i/f:value
			snprintf(path, sizeof(path), "/param/%s", iter->first.c_str());
			snprintf(types, sizeof(types), "%c", iter->second);
			
			lo_server_add_method(serv, path, types, ControlOSC::_param_set_handler, this);
			lo_server_add_method(serv, path, "ss", ControlOSC::_param_get_handler, this);		
			lo_server_add_method(serv, path, "s", ControlOSC::_param_get_handler, this);

			// cerr << "adding " << path << "  " << types << endl;
		}

		// readonly params
		const ParamMap & ro_param_map = _engine->get_ro_parameter_map();

		for (ParamMap::const_iterator iter = ro_param_map.begin();
		     iter != ro_param_map.end(); ++iter)
		{
			// get parameter s:retpath [s:returl]
			snprintf(path, sizeof(path), "/param/%s", iter->first.c_str());
			snprintf(types, sizeof(types), "%c", iter->second);
			
			lo_server_add_method(serv, path, "ss", ControlOSC::_param_get_handler, this);		
			lo_server_add_method(serv, path, "s", ControlOSC::_param_get_handler, this);

			// cerr << "adding ro " << path << "  " << types << endl;
		}

	}
}

bool
ControlOSC::init_osc_thread ()
{
	// create new thread to run server
	if (pipe (_request_pipe)) {
		cerr << "Cannot create osc request signal pipe" <<  strerror (errno) << endl;
		return false;
	}

	if (fcntl (_request_pipe[0], F_SETFL, O_NONBLOCK)) {
		cerr << "osc: cannot set O_NONBLOCK on signal read pipe " << strerror (errno) << endl;
		return false;
	}

	if (fcntl (_request_pipe[1], F_SETFL, O_NONBLOCK)) {
		cerr << "osc: cannot set O_NONBLOCK on signal write pipe " << strerror (errno) << endl;
		return false;
	}
	
	// cerr << "creating osc thread" << endl;
	OscThread worker(this);
	_osc_thread = new boost::thread(worker);

	
	//pthread_create (&_osc_thread, NULL, &ControlOSC::_osc_receiver, this);
	if (!_osc_thread) {
		return false;
	}
	//pthread_detach (_osc_thread);
	return true;
}

void
ControlOSC::terminate_osc_thread ()
{
	void* status;

	_shutdown = true;

	if (_osc_thread) {
		poke_osc_thread ();
		
		//pthread_join (_osc_thread, &status);
		_osc_thread->join();
		delete _osc_thread;
		_osc_thread = 0;
	}
}

void
ControlOSC::poke_osc_thread ()
{
	char c = 0;

	if (write (_request_pipe[1], &c, 1) != 1) {
		cerr << "cannot send signal to osc thread! " <<  strerror (errno) << endl;
	}
}



std::string
ControlOSC::get_server_url()
{
	string url;
	char * urlstr;

	if (_osc_server) {
		urlstr = lo_server_get_url (_osc_server);
		url = urlstr;
		free (urlstr);
	}
	
	return url;
}

std::string
ControlOSC::get_unix_server_url()
{
	string url;
	char * urlstr;

	if (_osc_unix_server) {
		urlstr = lo_server_get_url (_osc_unix_server);
		url = urlstr;
		free (urlstr);
	}
	
	return url;
}


/* server thread */

void
ControlOSC::osc_receiver()
{
	struct pollfd pfd[3];
	int fds[3];
	lo_server srvs[3];
	int nfds = 0;
	int timeout = 100; // time in ms to autoupdate
	int ret;
	
	fds[0] = _request_pipe[0];
	nfds++;
	
	if (_osc_server && lo_server_get_socket_fd(_osc_server) >= 0) {
		fds[nfds] = lo_server_get_socket_fd(_osc_server);
		srvs[nfds] = _osc_server;
		nfds++;
	}

	if (_osc_unix_server && lo_server_get_socket_fd(_osc_unix_server) >= 0) {
		fds[nfds] = lo_server_get_socket_fd(_osc_unix_server);
		srvs[nfds] = _osc_unix_server;
		nfds++;
	}
	
	
	while (!_shutdown) {

		for (int i=0; i < nfds; ++i) {
			pfd[i].fd = fds[i];
			pfd[i].events = POLLIN|POLLPRI|POLLHUP|POLLERR;
			pfd[i].revents = 0;
		}
		
	again:
		//cerr << "poll on " << nfds << " for " << timeout << endl;
		if ((ret = poll (pfd, nfds, timeout)) < 0) {
			if (errno == EINTR) {
				/* gdb at work, perhaps */
				cerr << "EINTR hit " << endl;
				goto again;
			}
			
			cerr << "OSC thread poll failed: " <<  strerror (errno) << endl;
			
			break;
		}
		
		if (ret == 0) {
			// timeout, lets autoupdate registered ro_values
			send_auto_updates();
		}

		// cerr << "poll returned " << ret << "  pfd[0].revents = " << pfd[0].revents << "  pfd[1].revents = " << pfd[1].revents << endl;
		
		if (_shutdown) {
			break;
		}
		
		if ((pfd[0].revents & ~POLLIN)) {
			cerr << "OSC: error polling extra port" << endl;
			break;
		}
		
		for (int i=1; i < nfds; ++i) {
			if (pfd[i].revents & POLLIN)
			{
				// this invokes callbacks
				//cerr << "invoking recv on " << pfd[i].fd << endl;
				lo_server_recv(srvs[i]);
			}
		}

	}

	// cerr << "OSC shutdown" << endl;
	
	if (_osc_server) {
		int fd = lo_server_get_socket_fd(_osc_server);
		if (fd >=0) {
				// hack around
			close(fd);
		}
		lo_server_free (_osc_server);
		_osc_server = 0;
	}

	if (_osc_unix_server) {
		cerr << "freeing unix server" << endl;
		lo_server_free (_osc_unix_server);
		_osc_unix_server = 0;
	}
	
	close(_request_pipe[0]);
	close(_request_pipe[1]);
}



/* STATIC callbacks */


int ControlOSC::_dummy_handler(const char *path, const char *types, lo_arg **argv, int argc,
			 void *data, void *user_data)
{
#ifdef DEBUG
	cerr << "got path: " << path << endl;
#endif
	return 0;
}


int ControlOSC::_quit_handler(const char *path, const char *types, lo_arg **argv, int argc,
			 void *data, void *user_data)
{
	ControlOSC * osc = static_cast<ControlOSC*> (user_data);
	return osc->quit_handler (path, types, argv, argc, data);

}

int ControlOSC::_ping_handler(const char *path, const char *types, lo_arg **argv, int argc,
			 void *data, void *user_data)
{
	ControlOSC * osc = static_cast<ControlOSC*> (user_data);
	return osc->ping_handler (path, types, argv, argc, data);

}

int ControlOSC::_param_set_handler(const char *path, const char *types, lo_arg **argv, int argc,
			 void *data, void *user_data)
{
	ControlOSC * osc = static_cast<ControlOSC*> (user_data);
	return osc->param_set_handler (path, types, argv, argc, data);

}
int ControlOSC::_param_get_handler(const char *path, const char *types, lo_arg **argv, int argc,
			 void *data, void *user_data)
{
	ControlOSC * osc = static_cast<ControlOSC*> (user_data);
	return osc->param_get_handler (path, types, argv, argc, data);

}

int ControlOSC::_param_get_all_handler(const char *path, const char *types, lo_arg **argv, int argc,
			 void *data, void *user_data)
{
	ControlOSC * osc = static_cast<ControlOSC*> (user_data);
	return osc->param_get_all_handler (path, types, argv, argc, data);

}


int ControlOSC::_register_update_handler(const char *path, const char *types, lo_arg **argv, int argc,
			 void *data, void *user_data)
{
	ControlOSC * osc = static_cast<ControlOSC*> (user_data);
	return osc->register_update_handler (path, types, argv, argc, data);
}

int ControlOSC::_unregister_update_handler(const char *path, const char *types, lo_arg **argv, int argc,
			 void *data, void *user_data)
{
	ControlOSC * osc = static_cast<ControlOSC*> (user_data);
	return osc->unregister_update_handler (path, types, argv, argc, data);
}



/* real callbacks */

int ControlOSC::quit_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data)
{
	_engine->quit();
	return 0;
}


int ControlOSC::ping_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data)
{
	string retpath;
	lo_address retaddr;
	int lentypes = strlen(types);

	if (lentypes == 0) {
		retpath = "/pingack";
	}
	else {
		retpath = &argv[0]->s;
	}

	if (lentypes > 1) {
		// return URL was given, use it
		string returl (&argv[1]->s);
		// cerr << "returl passed is: " << returl << endl;
		retaddr = find_or_cache_addr (returl);
		if (!retaddr) {
			cerr << "passed return URL is invalid" << endl;
			return 0;
		}
	}
	else {
		// try to use return address from sender
		lo_message msg = (lo_message) data;
		lo_address srcaddr = lo_message_get_source (msg);
		const char * sport = lo_address_get_port(srcaddr);
		int srcport = atoi(sport);

		if (srcaddr) {
			retaddr = srcaddr;
		} else {
			cerr << "source address is invalid" << endl;
			return 0;
		}
	}
	
	send_pingack (true, retaddr, retpath);

	return 0;
}


int ControlOSC::param_get_handler(const char *path, const char *types, lo_arg **argv, int argc,void *data)
{
	// s:retpath  [s:returl]
	string retpath (&argv[0]->s);
	string spath(path);
	lo_address retaddr;

	if (strlen(types) > 1) {
		// return URL was given, use it
		string returl (&argv[1]->s);
		// cerr << "returl passed is: " << returl << endl;
		retaddr = find_or_cache_addr (returl);
		if (!retaddr) {
			cerr << "passed return URL is invalid" << endl;
			return 0;
		}
	}
	else {
		lo_message msg = (lo_message) data;
		lo_address srcaddr = lo_message_get_source (msg);
		const char * sport = lo_address_get_port(srcaddr);
		int srcport = atoi(sport);
	
		// try to use return address from sender
		if (srcaddr) {
			retaddr = srcaddr;
		} else {
			cerr << "source address is invalid" << endl;
			return 0;
		}
	}

	// get value
	string param = spath.substr(7); // length of "/param/"

	AddressList retaddrs;
	retaddrs.push_back(AddrPathPair(retaddr,retpath));
	
	int rc = send_param_value (retaddrs, param);

	if (rc < 0) {
		fprintf(stderr, "OSC error %d: %s\n", lo_address_errno(retaddr), lo_address_errstr(retaddr));
	}

	return 0;
}

int ControlOSC::param_get_all_handler(const char *path, const char *types, lo_arg **argv, int argc,void *data)
{
	// s:retpath  [s:returl]
	string retpath (&argv[0]->s);
	string spath(path);
	lo_address retaddr;

	if (strlen(types) > 1) {
		// return URL was given, use it
		string returl (&argv[1]->s);
		// cerr << "returl passed is: " << returl << endl;
		retaddr = find_or_cache_addr (returl);
		if (!retaddr) {
			cerr << "passed return URL is invalid" << endl;
			return 0;
		}
	}
	else {
		lo_message msg = (lo_message) data;
		lo_address srcaddr = lo_message_get_source (msg);
		const char * sport = lo_address_get_port(srcaddr);
		int srcport = atoi(sport);
	
		// try to use return address from sender
		if (srcaddr) {
			retaddr = srcaddr;
		} else {
			cerr << "source address is invalid" << endl;
			return 0;
		}
	}

	AddressList retaddrs;
	retaddrs.push_back(AddrPathPair(retaddr,retpath));

	// for all params
	const ParamMap & param_map = _engine->get_parameter_map();
	
	for (ParamMap::const_iterator iter = param_map.begin();
	     iter != param_map.end(); ++iter)
	{
		string param = iter->first;

		int rc = send_param_value (retaddrs, param);
		
		if (rc < 0) {
			fprintf(stderr, "OSC error %d: %s\n", lo_address_errno(retaddr), lo_address_errstr(retaddr));
		}
	}

		// for all params
	const ParamMap & ro_param_map = _engine->get_parameter_map();
	
	for (ParamMap::const_iterator iter = ro_param_map.begin();
	     iter != ro_param_map.end(); ++iter)
	{
		string param = iter->first;

		int rc = send_param_value (retaddrs, param);
		
		if (rc < 0) {
			fprintf(stderr, "OSC error %d: %s\n", lo_address_errno(retaddr), lo_address_errstr(retaddr));
		}
	}

	
	return 0;
}


int ControlOSC::send_param_value(AddressList & retaddrs, const string & param, bool force)
{
	const ParamMap & param_map = _engine->get_parameter_map();
	const ParamMap & ro_param_map = _engine->get_ro_parameter_map();
	ParamMap::const_iterator found;
	char type;
	float fval;
	double dval;
	int ival;
	int64_t i64val;
	string sval;
	int rc = -2;

	if ((found = param_map.find(param)) != param_map.end()) {
		type = found->second;
	}
	else if ((found = ro_param_map.find(param)) != ro_param_map.end()) {
		type = found->second;
	}
	else {
		cerr << "could not find parameter: " << param << endl;
		return -1;
	}
	switch(type) {
	case 'i':
		if (_engine->get_param (param, ival) && (force || _state_update_i.check_param_updated (param, ival))) {
			for (AddressList::iterator iter = retaddrs.begin(); iter != retaddrs.end(); ++iter) {
				rc = lo_send (iter->first, iter->second.c_str(), "si", param.c_str(), ival);
			}
		}
		break;
	case 'h':
		if (_engine->get_param (param, i64val) && (force || _state_update_h.check_param_updated (param, i64val))) {
			for (AddressList::iterator iter = retaddrs.begin(); iter != retaddrs.end(); ++iter) {
				rc = lo_send (iter->first, iter->second.c_str(), "sh", param.c_str(), i64val);
			}
		}
		break;
	case 'f':
		if (_engine->get_param (param, fval) && (force || _state_update_f.check_param_updated (param, fval))) {
			for (AddressList::iterator iter = retaddrs.begin(); iter != retaddrs.end(); ++iter) {
				rc = lo_send (iter->first, iter->second.c_str(), "sf", param.c_str(), fval);
			}
		}
		break;
	case 'd':
		if (_engine->get_param (param, dval)  && (force || _state_update_d.check_param_updated (param, dval))) {
			for (AddressList::iterator iter = retaddrs.begin(); iter != retaddrs.end(); ++iter) {
				rc = lo_send (iter->first, iter->second.c_str(), "sd", param.c_str(), dval);
			}
		}
		break;
	case 's':
		if (_engine->get_param (param, sval) && (force || _state_update_s.check_param_updated (param, sval))) {
			for (AddressList::iterator iter = retaddrs.begin(); iter != retaddrs.end(); ++iter) {
				rc = lo_send (iter->first, iter->second.c_str(), "ss", param.c_str(), sval.c_str());
			}
		}
		break;

	default:
		break;
	}	

	//if (rc != -2) 
	//	cerr << "sending param: " << param << "  type: " << type << endl;
		

	return rc;
}

int ControlOSC::param_set_handler(const char *path, const char *types, lo_arg **argv, int argc,void *data)
{
	// /param/PARAM  f/i/s/h/d:val
	string spath(path);
	string param;
	float fval  = argv[0]->f;
	double dval  = argv[0]->d;
	int ival  = argv[0]->i;
	int64_t i64val  = argv[0]->h;
	string sval;

	param = spath.substr(7); // length of "/param/"
	// cerr << "types: " << types << "  param: " << param << endl;
	switch(types[0]) {
	case 'i':
		_engine->set_param (param, ival);
		break;
	case 'h':
		_engine->set_param (param, i64val);
		break;
	case 'f':
		_engine->set_param (param, fval);
		break;
	case 'd':
		_engine->set_param (param, dval);
		break;
	case 's':
		sval = &argv[0]->s;
		_engine->set_param (param, sval);
		// cerr << "setting " << param << " to " << sval << endl;
		break;

	default:
		break;
	}


	// cerr << "set handler: " << path << " param: " << param << endl;

	// send out updates to registered clients

	update_registered_param (param);

	//_engine->push_nonrt_event ( new ConfigUpdateEvent (ConfigUpdateEvent::Send, -2, _cmd_map->to_control_t(param), "", "", val, srcport));
	
	return 0;
}

void ControlOSC::update_registered_param (const std::string & param)
{
	send_param_value (_update_registrations, param, false);

}

int
ControlOSC::register_update_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data)
{
	// un/register_update args=s:retpath [s:returl]
	string returl;
	string retpath (&argv[0]->s);
	int lentypes = strlen(types);
	lo_address retaddr = 0;

	if (lentypes > 1) {
		// return URL was given, use it
		string returl (&argv[1]->s);
		// cerr << "returl passed is: " << returl << endl;
		retaddr = find_or_cache_addr (returl);
		if (!retaddr) {
			cerr << "passed return URL is invalid" << endl;
			return 0;
		}
	}
	else {
		// try to use return address from sender
		lo_message msg = (lo_message) data;
		lo_address srcaddr = lo_message_get_source (msg);
		const char * sport = lo_address_get_port(srcaddr);
		int srcport = atoi(sport);

		if (srcaddr) {
			retaddr = srcaddr;
		} else {
			cerr << "source address is invalid" << endl;
			return 0;
		}
	}

	AddrPathPair apair(retaddr, retpath);
	AddressList::iterator iter = find (_update_registrations.begin(), _update_registrations.end(), apair);

	if (iter == _update_registrations.end()) {
		_update_registrations.push_back (apair);
	}	

	//cerr << "register update handler " << endl;
	return 0;
}

int
ControlOSC::unregister_update_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data)
{
	// un/register_update args= s:retpath [s:returl]
	string returl;
	string retpath (&argv[0]->s);
	int lentypes = strlen(types);
	lo_address retaddr = 0;

	if (lentypes > 1) {
		// return URL was given, use it
		string returl (&argv[1]->s);
		//cerr << "returl passed is: " << returl << endl;
		retaddr = find_or_cache_addr (returl);
		if (!retaddr) {
			cerr << "passed return URL is invalid" << endl;
			return 0;
		}
	}
	else {
		// try to use return address from sender
		lo_message msg = (lo_message) data;
		lo_address srcaddr = lo_message_get_source (msg);
		const char * sport = lo_address_get_port(srcaddr);
		int srcport = atoi(sport);

		if (srcaddr) {
			retaddr = srcaddr;
		} else {
			cerr << "source address is invalid" << endl;
			return 0;
		}
	}

	AddrPathPair apair(retaddr, retpath);
	AddressList::iterator iter = find (_update_registrations.begin(), _update_registrations.end(), apair);

	if (iter != _update_registrations.end()) {
		//cerr << "erasing registration" << endl;
		_update_registrations.erase(iter);
	}	
	return 0;
}


lo_address
ControlOSC::find_or_cache_addr(string returl)
{
	lo_address addr = 0;

	if (returl.empty()) return 0;
	
	if (_retaddr_map.find(returl) == _retaddr_map.end()) {
		addr = lo_address_new_from_url (returl.c_str());
		if (addr && lo_address_errno (addr) < 0) {
			fprintf(stderr, "addr error %d: %s\n", lo_address_errno(addr), lo_address_errstr(addr));
		}
		_retaddr_map[returl] = addr;
	}
	else {
		addr = _retaddr_map[returl];
	}
	
	return addr;
}


void ControlOSC::send_auto_updates ()
{
	const ParamMap & ro_param_map = _engine->get_ro_parameter_map();

	for(ParamMap::const_iterator iter = ro_param_map.begin(); iter != ro_param_map.end(); ++iter) 
	{
		update_registered_param (iter->first);
	}
		// cerr << "ctrl " << ipair.second << " is new: " << val << endl;
}




void ControlOSC::send_all_config ()
{
	// for now just send pingacks to all registered addresses
	for (AddressList::iterator iter = _update_registrations.begin(); iter != _update_registrations.end(); ++iter)
	{
		send_pingack (true, (*iter).first, (*iter).second);
	}
}

void ControlOSC::send_error (std::string returl, std::string retpath, std::string mesg)
{
	lo_address addr;

	addr = find_or_cache_addr (returl);
	if (!addr) {
		return;
	}

	string oururl = get_server_url();
	
	if (lo_send(addr, retpath.c_str(), "ss", oururl.c_str(), mesg.c_str()) < 0) {
		fprintf(stderr, "OSC error %d: %s\n", lo_address_errno(addr), lo_address_errstr(addr));
	}
}

void ControlOSC::send_pingack (bool useudp, string returl, string retpath)
{
	lo_address addr;

	addr = find_or_cache_addr (returl);
	send_pingack(useudp, addr, retpath);
}

void ControlOSC::send_pingack (bool useudp, lo_address addr, string retpath)
{
	string oururl;

	if (!addr) {
		return;
	}

	if (!useudp) {
		oururl = get_unix_server_url();
	}

	// default to udp
	if (oururl.empty()) {
		oururl = get_server_url();
	}
	
	// cerr << "ac3jack: sending ping response" << endl;

	if (lo_send(addr, retpath.c_str(), "ss", oururl.c_str(), ac3jack_version) < 0) {
		fprintf(stderr, "OSC error %d: %s\n", lo_address_errno(addr), lo_address_errstr(addr));
	}
}


	
