/*
** Copyright (C) 2007 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#ifndef __ac3jack_control_osc__
#define __ac3jack_control_osc__


#include <lo/lo.h>
#include <string>
#include <map>
#include <list>
#include <utility>
#include "boost/thread/thread.hpp"

//#include <sigc++/object.h>


namespace AC3JACK {

class Engine;
	
	template <typename T> class StateUpdater
	{
	public:
		bool check_param_updated (const std::string & param, T const & value);
	protected:

		typedef std::map<std::string, T> ValueMap;
		
		ValueMap     _value_map;
	};



class ControlOSC
{
  public:
	
	ControlOSC (Engine *, unsigned int port);
	virtual ~ControlOSC();

	void start();
	
	std::string get_server_url();
	int get_server_port () { return _port; }

	std::string get_unix_server_url();
	
	bool is_ok() { return _ok; }

	void send_all_config ();
	void send_pingack (bool useudp, std::string returl, std::string retpath="/pingack");
	
	void send_auto_updates ();
	void send_error (std::string returl, std::string retpath, std::string mesg);
	void update_registered_param (const std::string & param);

  private:

	typedef std::pair<lo_address, std::string> AddrPathPair;
	typedef std::list<AddrPathPair> AddressList;
	
	void register_callbacks();
	
	lo_address find_or_cache_addr(std::string returl);
	void send_pingack (bool useudp, lo_address addr, std::string retpath="/pingack");
	
	int send_param_value (AddressList & retaddrs, const std::string & param, bool force=true);

	
	static int _quit_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
	static int _param_set_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
	static int _param_get_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
	static int _param_get_all_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
	static int _dummy_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
	static int _register_update_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
	static int _unregister_update_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
	static int _ping_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data, void *user_data);
	
	bool init_osc_thread();
	void terminate_osc_thread();
	void poke_osc_thread();

	class OscThread {
	   public:
                OscThread(ControlOSC * osc) : parent(osc) {}
                void operator()() { parent->osc_receiver(); }
           protected:
                ControlOSC * parent;
	};

	void osc_receiver();
	
	int quit_handler(const char *path, const char *types, lo_arg **argv, int argc,void *data);
	int ping_handler(const char *path, const char *types, lo_arg **argv, int argc,void *data);
	int param_get_all_handler(const char *path, const char *types, lo_arg **argv, int argc,void *data);
	int param_get_handler(const char *path, const char *types, lo_arg **argv, int argc,void *data);
	int param_set_handler(const char *path, const char *types, lo_arg **argv, int argc,void *data);
	int register_update_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data);
	int unregister_update_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data);
	

	Engine * _engine;

	boost::thread   * _osc_thread;
	int       _request_pipe[2];
	
	lo_server _osc_server;
	lo_server _osc_unix_server;
	std::string _osc_unix_socket_path;
	
	int _port;
	volatile bool _ok;
	volatile bool _shutdown;
	int _max_instance;
	
	std::map<std::string, lo_address> _retaddr_map;

	StateUpdater<double> _state_update_d;
	StateUpdater<float> _state_update_f;
	StateUpdater<int> _state_update_i;
	StateUpdater<int64_t> _state_update_h;
	StateUpdater<std::string> _state_update_s;
	
	AddressList _update_registrations;
};


template <typename T> bool 
StateUpdater<T>::check_param_updated (const std::string & param, T const & value)
{
	if (_value_map.find(param) != _value_map.end()) {
		//if (value == (*lastval).second) {
		if (value == _value_map.find(param)->second) {
			// same as last update, don't send
			return false;
		}
	}

	_value_map[param] = value;
	return true;
}
		


};




#endif
