/*
** Copyright (C) 2006 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#include <iostream>
#include <sstream>

#include <unistd.h>

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "engine.hpp"
#include "encoder.hpp"
#include "file_writer.hpp"
#include "control_osc.hpp"

#ifdef WITH_ALSA
#include "alsa_spdif_writer.hpp"
#endif

#ifdef WITH_COREAUDIO
#include "coreaudio_spdif_writer.hpp"
#endif


using namespace AC3JACK;
using namespace std;
using namespace PBD;
using namespace SigC;

#define MAX_CHANNELS 6
#define DISK_OUT_RB_SIZE 65536
#define SPDIF_OUT_RB_SIZE 16384
#define SPDIF_HEADER_SIZE 8
#define SAMPLE_MAX_16BIT  32768.0f
#define SAMPLE_MAX_24BIT  8388608.0f

#define SPDIF_FRAME_SIZE 6144

Engine::Engine()
{
	_ok = false;
	_driver = 0;
	_alive = true;
	_ac3buf = 0;
	_buffersize = 0;
	_frame_pos = 0;
	_sample_buffer = 0;
	_outsize = 0;
	_rawout_rb = 0;
	_transport_rolling = false;
	_curr_file_duration = 0.0;
	_curr_file_bytes = 0;
	_osc_control = 0;
	_encoder = 0;
	_file_writer = 0;
	_spdif_writer = 0;
	_conn_changed = false;
	
	// current parameters (r/w)
	_param_map["channels"] = 'i';
	_param_map["lfe"] = 'i';
	_param_map["duration"] = 'd';
	_param_map["spdif_out"] = 'i';
	_param_map["spdif_card"] = 's';
	_param_map["spdif_pcm"] = 'i';
	_param_map["jack_out"] = 'i';
	_param_map["jack_transport"] = 'i';	
	_param_map["outfile_name"] = 's';
	_param_map["file_output"] = 'i';
	_param_map["encoder/bitrate"] = 'i';
	_param_map["input_connections/left"] = 's';
	_param_map["input_connections/right"] = 's';
	_param_map["input_connections/center"] = 's';
	_param_map["input_connections/left_surround"] = 's';
	_param_map["input_connections/right_surround"] = 's';
	_param_map["input_connections/lfe"] = 's';	

	// ro, but not auto updating
	_param_map["spdif_devices"] = 's'; 
	_param_map["spdif_status"] = 's';
	_param_map["input_connectable_ports"] = 's';
	
	// read-only params
	_ro_param_map["curr_file_duration"] = 'd';
	_ro_param_map["curr_file_size"] = 'h';
}

Engine::~Engine()
{
	cleanup();
}


bool Engine::initialize(AudioDriver * driver)
{

	_driver = driver;
	_driver->set_engine(this);
	
	if (!_driver->initialize()) {
		cerr << "cannot connect to audio driver" << endl;
		return false;
	}

	_buffersize = _driver->get_buffersize();

	_driver->ConnectionsChanged.connect (slot (*this,  &Engine::on_driver_connections_changed));
	_driver->ConnectablesChanged.connect (slot (*this,  &Engine::on_driver_connectables_changed));	
	
	
	// create frame buffers for max channels
	/*
	for (size_t i = 0; i < MAX_CHANNELS; ++i) {
		sample_t * inbuf = new float[A52_SAMPLES_PER_FRAME];
		memset(inbuf, 0, sizeof(float) * A52_SAMPLES_PER_FRAME);
		_frame_buffers.push_back(inbuf);
	}
	*/

	// create interleaved framebuffer for MAX_CHANNELS
	_sample_buffer = new float[MAX_CHANNELS * A52_SAMPLES_PER_FRAME];
	

	// create ac3 buffer
	_ac3buf = new unsigned char[A52_MAX_CODED_FRAME_SIZE];
	memset(_ac3buf, 0, A52_MAX_CODED_FRAME_SIZE);

	_zerobuf = new unsigned char[SPDIF_FRAME_SIZE];
	memset(_zerobuf, 0, SPDIF_FRAME_SIZE);

	_rawout_rb = new RingBuffer<unsigned char>(32768);

	_samplerate = _driver->get_samplerate();
	_byterate = _samplerate * sizeof(short) * 2;


#ifdef WITH_ALSA
	_spdif_writer = new AlsaSpdifWriter(_driver->get_samplerate(), _params.spdif_dev);
#elif WITH_COREAUDIO
	// todo convert to id
	_spdif_writer = new CoreAudioSpdifWriter(_driver->get_samplerate(), _params.spdif_dev);
#else
	_last_spdif_status = "Not Available";
#endif

	setup_osc();
			
	reconfigure();

	_ok = true;

	return true;
}

bool
Engine::set_params(const EngineParams & params)
{
	_params = params;

	return true;
	//return reconfigure();
}

bool Engine::set_param (const std::string & name, int value)
{
	LockMonitor mon(_param_lock, __LINE__, __FILE__);

	if (name == "channels") {
		if (_params.channels != value) {
			_params.channels = value;
			setup_ports();
			setup_encoder();
			//cerr << "got new channels: " << value << endl;
		}
	}
	else if (name == "lfe") {
		if (_params.lfe != (value ? true : false)) {
			_params.lfe = (value ? true : false);
			setup_ports();
			setup_encoder();
			//cerr << "got new lfe: " << value << endl;
		}
	}
	else if (name == "spdif_out") {
		if (_params.spdif_out != (value ? true : false)) {
			_params.spdif_out =  value ? true : false;
			setup_spdif_out();
		}
	}
	else if (name == "spdif_pcm") {
		if (_params.spdif_pcm != (value ? true : false)) {
			_params.spdif_pcm =  value ? true : false;
			if (_spdif_writer) {
				_spdif_writer->set_allow_pcm (_params.spdif_pcm);
				_osc_control->update_registered_param("spdif_devices");
				if (!_params.spdif_pcm) {
					_params.spdif_dev = "";
					_osc_control->update_registered_param("spdif_card");
				}
				setup_spdif_out();
			}
		}
	}
	else if (name == "jack_transport") {
		_params.use_transport = value ? true : false;
	}
	else if (name == "jack_out") {
		if (_params.raw_jackout != (value ? true : false)) {
			_params.raw_jackout = value ? true : false;
			setup_ports();
		}
	}
	else if (name == "file_output") {
		if (_params.file_output != (value ? true : false)) {
			_params.file_output = value ? true : false;
			setup_file_out();
		}
	}
	else if (name == "encoder/bitrate") {
		if (_params.bitrate != value) {
			_params.bitrate = value;
			setup_encoder();
		}
	}
	else {
		return false;
	}

	return true;
}


bool Engine::get_param (const std::string & name, int & value)
{
	if (name == "channels") {
		value = _params.channels;
	}
	else if (name == "spdif_out") {
		value = _params.spdif_out ? 1 : 0;
	}
	else if (name == "spdif_pcm") {
		value = _params.spdif_pcm ? 1 : 0;
	}
	else if (name == "lfe") {
		value = _params.lfe ? 1 : 0;
	}
	else if (name == "file_output") {
		value = _params.file_output ? 1 : 0;
	}
	else if (name == "jack_transport") {
		value = _params.use_transport ? 1 : 0;
	}
	else if (name == "jack_out") {
		value = _params.raw_jackout ? 1 : 0;
	}
	else if (name == "encoder/bitrate") {
		value = _params.bitrate;
	}
	else {
		return false;
	}

	return true;	
}

bool Engine::set_param (const std::string & name, int64_t value)
{
	LockMonitor mon(_param_lock, __LINE__, __FILE__);

	return false;
}


bool Engine::get_param (const std::string & name, int64_t & value)
{
	if (name == "curr_file_size") {
		value = _curr_file_bytes;
	}
	else {
		return false;
	}

	return true;	
}


bool Engine::set_param (const std::string & name, float value)
{
	LockMonitor mon(_param_lock, __LINE__, __FILE__);

	//cerr << "set float param" << endl;
	return false;
}


bool Engine::get_param (const std::string & name, float & value)
{
	return false;
}

bool Engine::set_param (const std::string & name, double value)
{
	LockMonitor mon(_param_lock, __LINE__, __FILE__);

	//cerr << "set double param" << endl;
	if (name == "duration") {
		_params.duration = value;
	}
	else {
		return false;
	}
	return true;
}

bool Engine::get_param (const std::string & name, double & value)
{
	if (name == "duration") {
		value = _params.duration;
	}
	else if (name == "curr_file_duration") {
		value = _curr_file_duration;
	}
	else {
		return false;
	}
	return true;
}


bool Engine::set_param (const std::string & name, const std::string & value)
{
	LockMonitor mon(_param_lock, __LINE__, __FILE__);

	//cerr << "set string param: " << name << endl;
	if (name == "spdif_card") {
		if (_params.spdif_dev != value) {
			_params.spdif_dev = value;
			setup_spdif_out();
		}
	}
	else if (name == "outfile_name") {
		if (_params.outfile_name != value) {
			_params.outfile_name = value;
			setup_file_out();
		}
	}
	else if (name.find("input_connections") >= 0)
	{
		string channame = name.substr(name.find("/")+1);
		//cerr << "set channame is " << channame << endl;

		LockMonitor mon(_port_lock, __LINE__, __FILE__);

		connect_channel (channame, value);
	}
	else {
		return false;
	}
	return true;
}



bool Engine::get_param (const std::string & name, std::string & value)
{
	bool retval = true;
	
	if (name == "spdif_card") {
		value = _params.spdif_dev;
	}
	else if (name == "outfile_name") {
		value = _params.outfile_name;
	}
	else if (name == "spdif_status") {
		value = _last_spdif_status;
	}
	else if (name == "spdif_devices") {
		// value is multi-line string with tab delimited "Name \t devstring"
		DeviceInfoList infolist;
		if (_spdif_writer) {
			stringstream sstr;
			//cerr << "getting devices all the time" << endl;
			_spdif_writer->get_devices(infolist);
			for (DeviceInfoList::iterator iter = infolist.begin(); iter != infolist.end(); ++iter) {
				sstr << iter->first  << "\t" << iter->second << endl;
			}
			value = sstr.str();
			//cerr << "devices string is: " << endl << "'" << value << "'" << endl;
		}
		
	}
	else if (name == "input_connectable_ports")
	{
		AudioDriver::ConnectionList portlist;
		_driver->get_input_connectable_ports(portlist);

		stringstream sstr;
			
		for(AudioDriver::ConnectionList::iterator iter = portlist.begin(); iter != portlist.end(); ++iter)
		{
			sstr << *iter << endl;
		}
		value = sstr.str();
		//cerr << "connectable ports is:   " << value << endl;

	}
	else if (name.find("input_connections") >= 0)
	{
		AudioDriver::ConnectionList connlist;
		// lookup port number from name
		string channame = name.substr(name.find("/")+1);

		if (_name_portmap.find(channame) != _name_portmap.end())
		{
			LockMonitor mon(_port_lock, __LINE__, __FILE__);

			_driver->get_input_connections(_input_ports[_name_portmap[channame]], connlist);
			
			stringstream sstr;
			
			for(AudioDriver::ConnectionList::iterator iter = connlist.begin(); iter != connlist.end(); ++iter)
			{
				sstr << *iter << endl;
			}
			value = sstr.str();

			_chan_conns[channame] = value;
			
			//if (!value.empty())
			//	cerr << "connections to " << name << " are:   " << value << endl;
		}
		else {
			retval = false;
		}
	}
	else {
		retval = false;
	}
	
	return retval;
}


void Engine::connect_channel(const std::string & channame, const std::string &value)
{
	AudioDriver::ConnectionList connlist;
	//_driver->get_input_connections(_input_ports[_name_portmap[channame]], connlist);
	if (_name_portmap.find(channame) == _name_portmap.end())
	{
		return;
	}
	
	_driver->disconnect_input_port(_input_ports[_name_portmap[channame]], 0); // all

	// value is newline delimited connections to set
			
	int pos = 0;
	int endpos;

	if ((endpos = value.find("\n")) == -1) {
		endpos = value.size();
	}
			
	while (endpos > pos) {
		string line = value.substr(pos, endpos-pos);
		_driver->connect_input_port(_input_ports[_name_portmap[channame]], line.c_str()); 
		//cerr << "connecting " << line << " to " << channame << endl;

		if (endpos >= (int) value.size()) {
			break;
		}
				
		pos = endpos + 1;

		if ((endpos = value.find("\n", pos)) == -1) {
			endpos = value.size();
		}
	}
			
}


void
Engine::cleanup()
{
	if (!_ok) return;

	//cerr << "Engine::cleanup()" << endl;

	setup_ports (true); // cleanup only
	setup_osc (true);
	setup_file_out (true);
	setup_spdif_out (true);
	setup_encoder (true);

	LockMonitor mon(_port_lock, __LINE__, __FILE__);

	if (_spdif_writer) {
		delete _spdif_writer;
		_spdif_writer = 0;
	}

	delete [] _sample_buffer;

	delete [] _ac3buf;
	_ac3buf = 0;

	delete [] _zerobuf;
	_zerobuf = 0;
	
	delete _rawout_rb;
	_rawout_rb = 0;
	
	_ok = false;    
}

bool
Engine::reconfigure()
{
	setup_ports();

	setup_encoder();

	setup_file_out();

	setup_spdif_out(); 

	return true;
}

bool
Engine::setup_file_out(bool cleanup_only)
{
	LockMonitor mon(_port_lock, __LINE__, __FILE__);

	if (_file_writer) {
		delete _file_writer;
		_file_writer = 0;
	}
	if (cleanup_only) return true;

	_do_file_out = false;

	if (!_params.outfile_name.empty() && _params.file_output) {
		_file_writer = new FileWriter(_params.outfile_name, DISK_OUT_RB_SIZE);
		_file_writer->start();

		_do_file_out = true;
		_curr_file_duration = 0.0;
		_curr_file_bytes = 0;
		
		if (!_file_writer->ok()) {
			cerr << "Could not write to file: " << _params.outfile_name << endl;
			delete _file_writer;
			_file_writer = 0;
			_do_file_out = false;
		}
	}
	
	return _do_file_out;
}

bool
Engine::setup_osc(bool cleanup_only)
{
	if (_osc_control) {
		delete _osc_control;
		_osc_control = 0;
	}
	if (cleanup_only) return true;

	_osc_control = new ControlOSC (this, _params.osc_port);
	_osc_control->start();
	cerr << "OSC server at " << _osc_control->get_server_url() << endl;

	return true;
}

int Engine::get_osc_port()
{
	if (_osc_control) {
		return _osc_control->get_server_port();
	}
	return 0;
}

bool
Engine::setup_encoder(bool cleanup_only)
{
	LockMonitor mon(_port_lock, __LINE__, __FILE__);

	if (_encoder) {
		delete _encoder;
		_encoder = 0;
	}
	if (cleanup_only) return true;
	
	// create encoder
	_encoder = new Encoder();
	
	// initialize encoder
	EncoderParams enc_params;
	enc_params.channels = _params.channels;
	enc_params.sample_rate = _driver->get_samplerate();
	enc_params.bitrate = _params.bitrate;
	enc_params.lfe = _params.lfe;
	
	_encoder->set_params(enc_params);

	// todo, do a safe switch of the encoder context
	_encoder->initialize();

	return true;
}

bool
Engine::setup_spdif_out(bool cleanup_only)
{
	LockMonitor mon(_port_lock, __LINE__, __FILE__);

	_do_spdif_out = false;

	if (cleanup_only) return true;

	_last_spdif_status.clear();

	// spdif out
	if (_params.spdif_out) {

		if (_spdif_writer) {
			_spdif_writer->set_device (_params.spdif_dev);

			_spdif_writer->start();
			
			_do_spdif_out = true;
			
			_last_spdif_status = _spdif_writer->get_status();
			
			if (!_spdif_writer->ok()) {
				cerr << "Could not open spdif device: "  << _last_spdif_status << endl;
				//delete _spdif_writer;
				//_spdif_writer = 0;
				_do_spdif_out = false;
			}
		}
	} else if (_spdif_writer) {
		_spdif_writer->set_device("");
	}
	return true;
}


bool
Engine::setup_ports(bool cleanup_only)
{
	char portname[128];
	port_id_t tmpport;

	LockMonitor mon(_port_lock, __LINE__, __FILE__);

	// just remove them all
	
	for (size_t i=0; i < _input_ports.size(); ++i) 
	{
		_driver->destroy_input_port(_input_ports[i]);
	}
	
	_input_ports.clear();


	if (_params.raw_jackout && !cleanup_only) {
		if (_output_ports.size() == 0) {
			// only do once, always 2
			if (_driver->create_output_port ("nonaudio_ac3_1", tmpport)) {
				_output_ports.push_back (tmpport);
			}
			if (_driver->create_output_port ("nonaudio_ac3_2", tmpport)) {
				_output_ports.push_back (tmpport);
			}
		}
	}
	else {
		// destroy them
		for (size_t i=0; i < _output_ports.size(); ++i) 
		{
			_driver->destroy_output_port(_output_ports[i]);
		}
		_output_ports.clear();
	}

	_name_portmap.clear();
	
	if (cleanup_only) return true;
	/*
	 * 2 channels = L, R
	 * 3 channels = L, C, R
	 * 4 channels = L, R, LS, RS
	 * 5 ch       = L, C, R,  LS, RS
	 * 6 ch       = L, C, R,  LS, RS, LFE
	 */
	
	for (int i=0; i < _params.channels; ++i) 
	{
		switch(i)
		{
		case 0:
			strncpy (portname, "1_Left", sizeof(portname));
			_name_portmap["left"] = i;
			break;
			
		case 1:
			if (_params.channels == 2 || _params.channels == 4)
			{
				strncpy ( portname, "2_Right", sizeof(portname));
				_name_portmap["right"] = i;
			}
			else {
				strncpy ( portname, "2_Center", sizeof(portname));
				_name_portmap["center"] = i;
			}
		  
			break;
		case 2:
			if (_params.channels == 4) {
				strncpy (portname, "3_LeftSurround", sizeof(portname) );
				_name_portmap["left_surround"] = i;
			} else {
				strncpy (portname, "3_Right",  sizeof(portname));
				_name_portmap["right"] = i;
			}
			break;
		case 3:
			if (_params.channels == 4) {
				strncpy (portname, "4_RightSurround", sizeof(portname));
				_name_portmap["right_surround"] = i;
			}
			else {
				strncpy (portname, "4_LeftSurround", sizeof(portname));
				_name_portmap["left_surround"] = i;
			}
			break;
		case 4:
			if (_params.channels > 4) {
				strncpy (portname, "5_RightSurround", sizeof(portname));
				_name_portmap["right_surround"] = i;
			}
			break;
			//case 5:
			//strncpy (portname, "6_LFE", sizeof(portname));
			//break;
		default:
			break;
		}
		
		if (_driver->create_input_port (portname, tmpport)) {
			_input_ports.push_back (tmpport);
		}
	}

	if (_params.lfe) {
		snprintf (portname, sizeof(portname), "%d_LFE", _params.channels+1);

		if (_driver->create_input_port (portname, tmpport)) {
			_input_ports.push_back (tmpport);
		}
		_name_portmap["lfe"] = _input_ports.size()-1;		
	}

	// reconnect ports
	for (ChanConnMap::iterator iter=_chan_conns.begin(); iter != _chan_conns.end(); ++iter) {
		connect_channel (iter->first, iter->second);
	}
	
	/*
	for (i = 0; i < sources; i++) {
		if (i < nports && jack_connect (info->client, source_names[i], jack_port_name (ports[i]))) {
			fprintf (stderr, "cannot connect input port %s to %s\n", jack_port_name (ports[i]), source_names[i]);
		} 
	}
	*/
	
	return true;
}

void
Engine::buffersize_changed(unsigned int bufsize)
{

}


int
Engine::process(nframes_t nframes)
{
	// get input buffers
	sample_t * inbufs[MAX_CHANNELS];
	sample_t * outbufs[MAX_CHANNELS];
	TransportInfo trans_info;
	bool do_file_out = _do_file_out;
	
	TentativeLockMonitor mon(_port_lock, __LINE__, __FILE__);

	if (!mon.locked()) return 0;
	
	for (size_t i=0; i < _input_ports.size(); ++i) {
		inbufs[i] = _driver->get_input_port_buffer (_input_ports[i], _driver->get_buffersize());
	}

	for (size_t i=0; i < _output_ports.size(); ++i) {
		outbufs[i] = _driver->get_output_port_buffer (_output_ports[i], _driver->get_buffersize());
	}
	

	if (!_alive) {
		// silence on jack outs if necessary
		if (_params.raw_jackout) {
			for (size_t i=0; i < _output_ports.size(); ++i) {
				memset(outbufs[i], 0, nframes * sizeof(sample_t));
			}
		}
		return 0;
	}

	if (_params.use_transport && _file_writer && _params.file_output) {
		
		_driver->get_transport_info (trans_info);
		do_file_out = (trans_info.state == TransportInfo::ROLLING);
		// cerr << "transport " << trans_info.state << "  do file out: " << do_file_out << endl;
	}

	if (_params.duration > 0.0) {
		// only accurate to ac3 frame size right now
		do_file_out = do_file_out && (_curr_file_duration < _params.duration);
	}
	
	/* do the actual encoding here in the realtime thread.
	 * this is not ideal, because it makes small periods bursty
	 * we should do something like jamin
	 */
	
	// fill and process frame buffers as appropriate
	nframes_t frames_left = A52_SAMPLES_PER_FRAME - _frame_pos;
	nframes_t offset = 0;

	while (offset < nframes)
	{
		if ((nframes-offset) >= frames_left)
		{
			// copy only frames_left more data
			nframes_t pos = _frame_pos * _input_ports.size();
			for (nframes_t spos=offset;  spos < offset+frames_left; ++spos) {
				for (size_t i=0; i < _input_ports.size(); ++i) 
				{
					_sample_buffer[pos+i] = inbufs[i][spos];
				}
				pos += _input_ports.size();
			}  
			
			// use interleaved version
			_outsize = _encoder->run (_ac3buf + SPDIF_HEADER_SIZE, _sample_buffer, A52_SAMPLES_PER_FRAME);


			// push ac3 output to file writer thread (not including spdif header and pre-swab)
			if (do_file_out) {
				_file_writer->write(_ac3buf + SPDIF_HEADER_SIZE, _outsize);
				_curr_file_duration += A52_SAMPLES_PER_FRAME / (double) _driver->get_samplerate();
				_curr_file_bytes += _outsize;
				// cerr << "outputting: current duration: " << _curr_file_duration << endl;
			}
			
			fill_spdif_header (_ac3buf, _outsize + SPDIF_HEADER_SIZE);
			
			// push ac3 output to hw spdif output thread
			float calc_ac3byterate = (_outsize * _samplerate / (float) A52_SAMPLES_PER_FRAME);  
			nframes_t silencebytes = (nframes_t) (_outsize * (_byterate / calc_ac3byterate)) - _outsize - SPDIF_HEADER_SIZE;
			//cerr << "total byterate: " << _byterate << "  ac3 byterate is " << calc_ac3byterate 
			//    << " (bitrate " << calc_ac3byterate * 8 << " ) bytes: " << _outsize << " + " 
			//     << SPDIF_HEADER_SIZE << "   silence bytes=" << silencebytes << endl;

			if (_do_spdif_out && _spdif_writer) {
				_spdif_writer->write (_ac3buf, _outsize + SPDIF_HEADER_SIZE);
				// write the proper remainder of zero padding
				_spdif_writer->write(_zerobuf, silencebytes);
			}

			// pump raw spdif out to driver outputs if configured
			if (_params.raw_jackout)
			{
				_rawout_rb->write(_ac3buf, _outsize + SPDIF_HEADER_SIZE);

				// write the proper remainder of zero padding
				// inefficient, should be memsetting
				_rawout_rb->write(_zerobuf, silencebytes);
			}
			
			
			offset += frames_left;
			frames_left = A52_SAMPLES_PER_FRAME;
			_frame_pos = 0;
		}
		else
		{
			// copy incoming data into frame buffers without processing

			nframes_t pos = _frame_pos * _input_ports.size();
			for (nframes_t spos=offset;  spos < nframes; ++spos) {
				for (size_t i=0; i < _input_ports.size(); ++i) 
				{
					_sample_buffer[pos+i] = inbufs[i][spos];
				}
				pos += _input_ports.size();
			}  

			_frame_pos += (nframes-offset);
			offset += (nframes-offset);
		}
	}

	// pump raw spdif out to driver outputs if configured
	if (_params.raw_jackout)
	{
		output_to_driver (nframes);
	}
			

	
	return 0;
}

void
Engine::fill_spdif_header (unsigned char * buf, int outsize)
{
	// todo, use outsize and not assume the fixed frame size?
	int ac3outsize = outsize - SPDIF_HEADER_SIZE;
	
	buf[0] = 0x72; buf[1] = 0xf8;	/* spdif syncword */
	buf[2] = 0x1f; buf[3] = 0x4e;	/* .............. */
	buf[4] = 0x01;			/* AC3 data */
	buf[5] = buf[13] & 7;		/*  bsmod, stream = 0 */
	buf[6] = (ac3outsize << 3) & 0xff;
	buf[7] = (ac3outsize >> 5) & 0xff;

	//fprintf(stderr, "buf[8] = %02x  buf[9] = %02x\n", buf[8], buf[9]);
	//buf[8] = 0x77; buf[9] = 0x0b;	/* AC3 syncword ???*/
	//swab(buf+10, buf+10, outsize - 10);
	//fprintf(stderr, "ac3 outsize %d\n", ac3outsize);
	
	//swab(buf, buf, 8);

	// swab the data (should this only be done if the platform is little-endian??)
#if !IS_BIGENDIAN
	//swab(buf+SPDIF_HEADER_SIZE+2, buf+SPDIF_HEADER_SIZE+2, outsize - SPDIF_HEADER_SIZE - 2);
	swab(buf+SPDIF_HEADER_SIZE, buf+SPDIF_HEADER_SIZE, ac3outsize);
#endif

}

int
Engine::output_to_driver (nframes_t nframes)
{
	sample_t  * outbufs[2];
	int wrotebytes = 0;
	nframes_t nframes_left = nframes;
	
	if (_output_ports.size() < 2) return 0;
	
	outbufs[0] = _driver->get_output_port_buffer (_output_ports[0], _driver->get_buffersize());
	outbufs[1] = _driver->get_output_port_buffer (_output_ports[1], _driver->get_buffersize());
	
	if (_rawout_rb->read_space() == 0) {
		//if (_rawout_rb->read_space() < SPDIF_FRAME_SIZE) {
		// just write silence
		memset (outbufs[0], 0, nframes * sizeof(sample_t));
		memset (outbufs[1], 0, nframes * sizeof(sample_t));		
		// cerr << "writing " << nframes << " of silence no data" << endl;
	}
	else
	{
		RingBuffer<unsigned char>::rw_vector rvec;

		_rawout_rb->get_read_vector(&rvec);
		
		while (nframes_left > 0 && rvec.len[0] > 4) {

			
			nframes_t towrite_frames = (rvec.len[0]) / (sizeof(short) * 2);
			towrite_frames = std::min (towrite_frames, nframes_left);
			
			// write and deinterleave into the two channels
#if 1
			sample_move_dS_s16 (outbufs[0] + (nframes - nframes_left), (char *) rvec.buf[0], towrite_frames, sizeof(short) * 2);
			sample_move_dS_s16 (outbufs[1] + (nframes - nframes_left), (char *) rvec.buf[0] + sizeof(short), towrite_frames, sizeof(short) * 2);
#else
			sample_move_dS_s16_24ph (outbufs[0] + (nframes - nframes_left), (char *) rvec.buf[0], towrite_frames, sizeof(short) * 2);
			sample_move_dS_s16_24ph (outbufs[1] + (nframes - nframes_left), (char *) rvec.buf[0] + sizeof(short), towrite_frames, sizeof(short) * 2);
#endif			
			wrotebytes = towrite_frames * sizeof(short) * 2;
			nframes_left -= towrite_frames;
			
			_rawout_rb->increment_read_ptr (wrotebytes);
			_rawout_rb->get_read_vector(&rvec);
		}


		if (nframes_left > 0) {
			// write silence
			memset (outbufs[0] + (nframes - nframes_left), 0, (nframes_left) * sizeof(sample_t));
			memset (outbufs[1] + (nframes - nframes_left), 0, (nframes_left) * sizeof(sample_t));		
			// cerr << "writing " << nframes_left << " of silence" << endl;
		}
	}

	return wrotebytes;
}


void
Engine::sample_move_dS_s16 (sample_t *dst, char *src, nframes_t nsamples, unsigned long src_skip) 
{
	/* ALERT: signed sign-extension portability !!! */
	while (nsamples--) {
		*dst = (*((short *) src)) / SAMPLE_MAX_16BIT;
		dst++;
		src += src_skip;
	}
}	

void
Engine::sample_move_dS_s16_24ph (sample_t *dst, char *src, nframes_t nsamples, unsigned long src_skip) 
{
	/* ALERT: signed sign-extension portability !!! */
	while (nsamples--) {
		*dst = (((int)(*((short *) src))) << 8)/ SAMPLE_MAX_24BIT;
		dst++;
		src += src_skip;
	}
}	

void
Engine::mainloop()
{
	while (_alive)
	{
		usleep(250000);

		if (_params.duration > 0.0 && _curr_file_duration >= _params.duration) {
			// send message to disable file writing
			// just close out the file writer
			set_param ("file_output", (int) 0);
			_osc_control->update_registered_param("file_output");
		}
	}
}

void
Engine::idle_work()
{
	if (_alive)
	{
		if (_params.duration > 0.0 && _curr_file_duration >= _params.duration) {
			// send message to disable file writing
			// just close out the file writer
			set_param ("file_output", (int) 0);
			_osc_control->update_registered_param("file_output");
		}

		if (_conn_changed) {
			
			_osc_control->update_registered_param("input_connections/left");
			_osc_control->update_registered_param("input_connections/right");
			_osc_control->update_registered_param("input_connections/center");
			_osc_control->update_registered_param("input_connections/left_surround");
			_osc_control->update_registered_param("input_connections/right_surround");
			_osc_control->update_registered_param("input_connections/lfe");	
			//cerr << "sending conn CHANGED" << endl;
			_conn_changed = false;
		}
	}
}

void
Engine::quit()
{
	if (!_params.ignore_quit) {
		cleanup();
		_alive = false;
	}
}


void Engine::on_driver_connections_changed()
{
	//cerr << "connections changed" << endl;

	_conn_changed = true;
	
}

void Engine::on_driver_connectables_changed()
{
	_osc_control->update_registered_param("input_connectable_ports");
	
}
