/*
** Copyright (C) 2006 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#ifndef __ac3jack_engine__
#define __ac3jack_engine__

#include <vector>
#include <string>
#include <stdint.h>
#include <map>

#include "audio_driver.hpp"
#include "lockmonitor.hpp"
#include "ringbuffer.hpp"
#include "spdif_writer.hpp"

namespace AC3JACK
{

class Encoder;	
class FileWriter;
class ControlOSC;	

struct EngineParams
{
	EngineParams() :
		channels(5),duration(0),
		outfile_name(""), file_output(false), bitrate(448), use_transport(false), raw_jackout(false), 
		spdif_out(false),spdif_pcm(false), spdif_dev(""), 
		osc_port(12314), ignore_quit(false), lfe(false)
		{} 
		
	int    channels;
	double  duration;
	std::string outfile_name;
	bool   file_output;
	int    bitrate;
	bool   use_transport;
	bool   raw_jackout;
	bool        spdif_out;
	bool        spdif_pcm;
	std::string spdif_dev;
	std::vector<std::string> portconns;
	int    osc_port;
	bool   ignore_quit;
	bool   lfe;
};


// key is parameter name (osc path), value is type (s,i,f,d, etc)
typedef std::map<std::string, char> ParamMap;

class Engine
	: public SigC::Object
{
  public:

	Engine();
	virtual ~Engine();

	bool set_params(const EngineParams & params);
	
	bool set_param (const std::string & name, int value);
	bool set_param (const std::string & name, int64_t value);
	bool set_param (const std::string & name, float value);
	bool set_param (const std::string & name, double value);
	bool set_param (const std::string & name, const std::string & value);

	bool get_param (const std::string & name, int & value);
	bool get_param (const std::string & name, int64_t & value);
	bool get_param (const std::string & name, float & value);
	bool get_param (const std::string & name, double & value);
	bool get_param (const std::string & name, std::string & value);

	bool initialize(AudioDriver * driver);
	void cleanup();

	AudioDriver * get_audio_driver () { return _driver; }

	bool is_ok() const { return _ok; }

	double get_file_duration() const { return _curr_file_duration; }
	int64_t get_file_bytes() const { return _curr_file_bytes; }	

	const ParamMap & get_parameter_map() { return _param_map; }
	const ParamMap & get_ro_parameter_map() { return _ro_param_map; }
	
	int get_osc_port();
	
	int  process (nframes_t);
	void buffersize_changed (nframes_t);

	void mainloop();
	void idle_work();

	void quit();

	static void sample_move_dS_s16 (sample_t *dst, char *src, nframes_t nsamples, unsigned long src_skip); 
	static void sample_move_dS_s16_24ph (sample_t *dst, char *src, nframes_t nsamples, unsigned long src_skip); 
	
protected:

	bool reconfigure();
	bool setup_ports(bool cleanup_only=false);
	bool setup_osc(bool cleanup_only=false);
	bool setup_encoder(bool cleanup_only=false);
	bool setup_spdif_out(bool cleanup_only=false);
	bool setup_file_out(bool cleanup_only=false);

	void fill_spdif_header (unsigned char * buf, int outsize);
	//int output_to_driver (unsigned char * ac3buf, int bufsize, nframes_t nframes);
	int output_to_driver (nframes_t nframes);

	void on_driver_connections_changed();
	void on_driver_connectables_changed();	

	void connect_channel(const std::string & channame, const std::string &value);

	AudioDriver * _driver;
	volatile bool _ok;
	volatile bool _alive;
	volatile bool _conn_changed;
	nframes_t     _buffersize;

	EngineParams  _params;

	ParamMap      _param_map;
	ParamMap      _ro_param_map;
	
	PBD::NonBlockingLock    _port_lock;

	// only taken when doing set params
	PBD::NonBlockingLock    _param_lock;

	std::vector<port_id_t>  _input_ports;
	std::vector<port_id_t>  _output_ports;

	typedef std::map<std::string,port_id_t> NamePortMap;
	NamePortMap _name_portmap;

	typedef std::map<std::string,std::string> ChanConnMap;
	ChanConnMap _chan_conns;
	
	
	std::vector<sample_t *>   _frame_buffers;
	sample_t   *              _sample_buffer;

	RingBuffer<unsigned char>  * _rawout_rb;

	nframes_t               _frame_pos;
	
	unsigned char *         _ac3buf;
	unsigned char *         _zerobuf;

	int                     _outsize; // bytes

	FileWriter             * _file_writer;
	volatile double          _curr_file_duration;
	volatile int64_t         _curr_file_bytes;

        SpdifWriter            * _spdif_writer;

	ControlOSC             * _osc_control;
	
	Encoder  *              _encoder;

	bool                    _do_spdif_out;
	bool                    _do_file_out;
	bool                    _transport_rolling;
	std::string             _last_spdif_status;
	nframes_t               _samplerate;
	nframes_t               _byterate;
};


}

#endif
