/*
** Copyright (C) 2006 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#include <iostream>
#include "boost/date_time/posix_time/posix_time.hpp"
#include "boost/thread/thread.hpp"
#include "boost/thread/xtime.hpp"

#include "file_writer.hpp"


using namespace AC3JACK;
using namespace std;
using namespace boost;

FileWriter::FileWriter(const std::string & filename, size_t bufsize)
	: ThreadWriter(), _filename(filename), _outfile(0)
{
	_buffer = new RingBuffer<unsigned char>(bufsize);

	open_output_file();
}

FileWriter::~FileWriter()
{
	// TODO: ensure thread is done
	_active = false;

	{
		boost::try_mutex::scoped_lock lk(_butler_mutex);
		_more_incoming.notify_one();
	}

	cerr << "waiting for butler thread to end in filewriter" << endl;
	_butler_thread->join();
	cerr << "joined butler" << endl;

	if (_outfile) {
		fclose(_outfile);
		_outfile = 0;
	}
	
	delete _buffer;
}

bool FileWriter::open_output_file()
{
	boost::try_mutex::scoped_lock lk(_butler_mutex);

	if (_outfile) {
		fflush(_outfile);
		fclose(_outfile);
		_outfile = 0;
	}
	
	// create (try) output file
	cerr << "attempting to open" << endl;

	if (_filename == "stdout") {
		_outfile = stdout;
	} else {
		_outfile = fopen(_filename.c_str(), "w");
	}
	
	if (!_outfile) {
		cerr << "error opening " << _filename << " for writing" << endl;
		_ok = false;
	}
	_ok = true;

	return _ok;
}


size_t FileWriter::write(unsigned char * data, size_t length)
{
	size_t wrote;

	wrote = _buffer->write (data, length);

	// non blocking try lock
	boost::try_mutex::scoped_try_lock lk(_butler_mutex);
	if (lk) {
		_more_incoming.notify_one();
	}
					

	return wrote;
}

// write thread routines follow

void FileWriter::run_work_thread()
{
	const size_t MIN_BLOCKSIZE = 1024;
	
	// this is released when we are waiting for more to write
	boost::try_mutex::scoped_lock lk(_butler_mutex);

	cerr << "started filewriter work thread" << endl;
	
	while (_active)
	{
                // sleep till we are needed
                wait_for_more(lk);

                if (!_active) {
                        break;
                }

		write_to_disk (MIN_BLOCKSIZE);
		
	}

	// time to stop, make sure to write out any remaining stuff

	write_to_disk (0);

	cerr << "finished filewriter work thread" << endl;	
}

void FileWriter::write_to_disk(size_t minblocksize)
{
	RingBuffer<unsigned char>::rw_vector readvec;

	if (_outfile)
	{
		while (_buffer->read_space() > minblocksize) {
				
			_buffer->get_read_vector(&readvec);

			fwrite (readvec.buf[0], readvec.len[0], 1, _outfile);

			if (readvec.len[1] > 0) {
				fwrite (readvec.buf[1], readvec.len[1], 1, _outfile);
			}

			_buffer->increment_read_ptr (readvec.len[0] + readvec.len[1]);
		}

		// fflush(_outfile);
	}
}

void FileWriter::wait_for_more(try_mutex::scoped_lock & lk)
{
	// we just need to sleep until more data shows up
	// timeout just in case
	xtime sleeptime;

        while (_buffer->read_space() == 0 && _active) {
                xtime_get(&sleeptime, TIME_UTC);
                sleeptime.nsec += 100000000;
                if (sleeptime.nsec >= 1000000000) {
                        sleeptime.sec += 1;
                        sleeptime.nsec -= 1000000000;
                }
                _more_incoming.timed_wait (lk, sleeptime);
        }

}
