/*
** Copyright (C) 2004 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <signal.h>
#include <pthread.h>

#include <cstdio>
#include <iostream>
#include <cstring>

#ifdef __WXMAC__
#include <wx/filename.h>
wxString GetExecutablePath(wxString argv0);
#endif

//#include <wx/wx.h>

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

// for all others, include the necessary headers (this file is usually all you
// need because it includes almost all "standard" wxWindows headers
#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include <wx/cmdline.h>

#include "version.h"

#include "gui_app.hpp"
#include "app_frame.hpp"
#include "main_panel.hpp"
#include "osc_client.hpp"
#include "engine.hpp"
#include "jack_audio_driver.hpp"

using namespace AC3JACKGui;
using namespace AC3JACK;
using namespace std;

#define GUI_OSC_PORT 12333

enum {
	ID_IdleTimer = 9000
};

// Create a new application object: this macro will allow wxWindows to create
// the application object during program execution (it's better than using a
// static object for many reasons) and also declares the accessor function
// wxGetApp() which will return the reference of the right type (i.e. MyApp and
// not wxApp)
IMPLEMENT_APP(AC3JACKGui::GuiApp)


BEGIN_EVENT_TABLE(AC3JACKGui::GuiApp, wxApp)

	EVT_TIMER(ID_IdleTimer, GuiApp::on_idle_timer)
	
END_EVENT_TABLE()

	
// ============================================================================
// implementation
// ============================================================================

// ----------------------------------------------------------------------------
// the application class
// ----------------------------------------------------------------------------


static const wxCmdLineEntryDesc cmdLineDesc[] =
{
	{ wxCMD_LINE_SWITCH, wxT("h"), wxT("help"), wxT("show this help"), wxCMD_LINE_VAL_NONE, wxCMD_LINE_OPTION_HELP },
	{ wxCMD_LINE_SWITCH, wxT("V"), wxT("version"), wxT("show version only"), wxCMD_LINE_VAL_NONE },
	{ wxCMD_LINE_OPTION, wxT("H"), wxT("connect-host"), wxT("connect to remote ac3jack on given host")},
	{ wxCMD_LINE_OPTION, wxT("P"), wxT("connect-port"), wxT("connect to remote ac3jack on given port"), wxCMD_LINE_VAL_NUMBER },
	{ wxCMD_LINE_OPTION, wxT("J"), wxT("jack-name"), wxT("jack client name, default is ac3jack_1")},
	{ wxCMD_LINE_OPTION, wxT("S"), wxT("jack-server-name"), wxT("specify JACK server name")},
	{ wxCMD_LINE_NONE }
};
	

bool
GuiApp::parse_options (int argc, wxChar **argv)
{
	wxCmdLineParser parser(argc, argv);
	parser.SetDesc(cmdLineDesc);

	wxString logotext = wxT("AC3Jack ") +
		wxString::FromAscii (ac3jack_version) +
		wxT("\nCopyright 2007 Jesse Chappell\n")
		wxT("AC3Jack comes with ABSOLUTELY NO WARRANTY\n")
		wxT("This is free software, and you are welcome to redistribute it\n")
		wxT("under certain conditions; see the file COPYING for details\n");

	parser.SetLogo (logotext);


	int ret = parser.Parse();

	if (ret != 0) {
		// help or error
		return false;
	}

	wxString strval;
	long longval;

	if (parser.Found (wxT("V"))) {
		fprintf(stderr, "%s\n", (const char *) logotext.ToAscii());
		return false;
	}
	

	parser.Found (wxT("H"), &_host);

	if (parser.Found (wxT("P"), &longval)) {
		_port = longval;
	}

	parser.Found (wxT("S"), &_server_name);
	parser.Found (wxT("J"), &_client_name);

	return true;
}
	
	
GuiApp::GuiApp()
	: _frame(0), _host(wxT("")), _port(0)
{
	_show_usage = 0;
	_show_version = 0;
	_engine = 0;
}

GuiApp::~GuiApp()
{

	if (_engine) {
		AudioDriver * driver = _engine->get_audio_driver();
		driver->deactivate();
		usleep(200000);
		driver->set_engine(0);
		
		delete _engine;
		_engine = 0;
	
		delete driver;
	}
}


// `Main program' equivalent: the program execution "starts" here
bool GuiApp::OnInit()
{

	
	wxString jackname;
	wxString rcdir;
	wxString jackdir;
	
	SetExitOnFrameDelete(TRUE);

	
	// use stderr as log
	wxLog *logger=new wxLogStderr();
	logger->SetTimestamp(NULL);
	wxLog::SetActiveTarget(logger);
	
	
	if (!parse_options(argc, argv)) {
		// do not continue
		return FALSE;
	}


	// Create the main application window
	_frame = new AppFrame (wxT("AC3Jack"), wxPoint(100, 100), wxDefaultSize);


	
	// override defaults
	OscClient & osc_client = _frame->get_main_panel()->get_osc_control();

	if (!_host.empty()) {
		osc_client.set_host (_host);
	}
	if (_port != 0) {
		osc_client.set_port (_port);
	}

	osc_client.ConnectFailed.connect (slot (*this,  &GuiApp::on_connect_failed));
	
	
	/*
	if (!_host.empty()) {
		loopctrl.get_spawn_config().host = _host.ToAscii();
	}
	if (_port != 0) {
		loopctrl.get_spawn_config().port = _port;
	}
	if (!_client_name.empty()) {
		loopctrl.get_spawn_config().jack_name = _client_name.ToAscii();
	}
	if (!_server_name.empty()) {
		loopctrl.get_spawn_config().jack_serv_name = _server_name.ToAscii();
	}
	*/

	// connect
	osc_client.connect();
	

	// Show it and tell the application that it's our main window
	//_frame->SetSizeHints(500, 210);
	//_frame->SetSize(500, 215);
	_frame->Show(TRUE);

	SetTopWindow(_frame);

	_frame->Raise();
	
	// success: wxApp::OnRun() will be called which will enter the main message
	// loop and the application will run. If we returned FALSE here, the
	// application would exit immediately.
	return TRUE;
}


void 
GuiApp::on_connect_failed (const std::string & msg)
{
	//wxMessageDialog dial(this, wxString::FromAscii(msg.c_str()), wxT("Connection Error"), wxOK);
	//dial.SetTitle(wxT("Connection Error"));
	//dial.ShowModal();

	// start our own engine
	if (!_engine) {
		start_engine();
	}
	else {
		wxMessageDialog dial(_frame, wxString(wxT("Internal Engine start failure")), wxString(wxT("Internal Engine connect failure")), wxOK);
		dial.SetTitle(wxT("Engine connect failure"));
		dial.ShowModal();
	}
}

void GuiApp::start_engine()
{
	AudioDriver * driver = new JackAudioDriver((const char *)_client_name.ToAscii(), (const char *) _server_name.ToAscii());

	_engine = new Engine();

	EngineParams engine_params;
	engine_params.osc_port = GUI_OSC_PORT;
	engine_params.ignore_quit = true;

	// TODO load rc
	_engine->set_params (engine_params);
	
	if (!_engine->initialize(driver)) {
		wxMessageDialog dial(_frame, wxString(wxT("Engine start failure")), wxString(wxT("Engine start failure")), wxOK);
		dial.SetTitle(wxT("Engine start failure"));
		dial.ShowModal();
		return;
	}
	cerr << "initialized engine" << endl;
	
	if (!driver->activate()) {
		wxMessageDialog dial(_frame, wxString(wxT("Engine activate failure")), wxString(wxT("Engine activate failure")), wxOK);
		dial.SetTitle(wxT("Engine activate failure"));
		dial.ShowModal();
		return;
	}
	cerr << "activated  driver" << endl;
	
	OscClient & osc_client = _frame->get_main_panel()->get_osc_control();
	osc_client.set_host(wxT("127.0.0.1"));
	osc_client.set_port(_engine->get_osc_port());
	osc_client.connect();

	cerr << "connected" << endl;	

	_idle_timer = new wxTimer(this, ID_IdleTimer);
	_idle_timer->Start(1000, false);

	
}

void
GuiApp::on_idle_timer(wxTimerEvent &ev)
{
	if (_engine) {
		_engine->idle_work();
	}
}

#ifdef __WXMAC__
wxString
GetExecutablePath(wxString argv0)
{
    wxString path;

    if (wxIsAbsolutePath(argv0))
        path = argv0;
    else {
        wxPathList pathlist;
        pathlist.AddEnvList(wxT("PATH"));
        path = pathlist.FindAbsoluteValidPath(argv0);
    }

    wxFileName filename(path);
    filename.Normalize();

    path = filename.GetFullPath();
    path = path.BeforeLast('/') + wxT("/");

    return path;
}
#endif // __WXMAC__

