/*
** Copyright (C) 2004 Jesse Chappell <jesse@essej.net>
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/

#include <cstdio>
#include <iostream>
#include <cerrno>
#include "osc_client.hpp"

#include <wx/filename.h>

#include <sys/poll.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>

using namespace std;
using namespace AC3JACKGui;
//using namespace AC3JACK;

#define DEFAULT_OSC_PORT 12314


OscClient::OscClient (const wxString & rcdir)
{
	_osc_traffic_thread = 0;
	_osc_addr = 0;
	_rcdir = rcdir;
	_sentinel = true;
	_lastchance = false;
	_we_spawned = false;

	_port = DEFAULT_OSC_PORT;
	
	setup_param_map();
	
	_osc_server = lo_server_new(NULL, NULL);
	if (!_osc_server) return;
	
	char * tmpstr;
	tmpstr = lo_server_get_url (_osc_server);
	_our_url = tmpstr;
	free (tmpstr);

	cerr << "ac3jackgui: our URL is " << _our_url << endl;
	
	/* add handler for control param callbacks, first is loop index , 2nd arg ctrl string, 3nd arg value */
	lo_server_add_method(_osc_server, "/param", 0, OscClient::_param_handler, this);
	//lo_server_add_method(_osc_server, "/param", "sh", OscClient::_param_handler, this);
	//lo_server_add_method(_osc_server, "/param", "sd", OscClient::_param_handler, this);
	//lo_server_add_method(_osc_server, "/param", "sf", OscClient::_param_handler, this);
	//lo_server_add_method(_osc_server, "/param", "ss", OscClient::_param_handler, this);

	// pingack expects: s:engine_url s:version i:loopcount
	lo_server_add_method(_osc_server, "/pingack", "ss", OscClient::_pingack_handler, this);

	lo_server_add_method(_osc_server, "/alive_resp", "ss", OscClient::_alive_handler, this);

	lo_server_add_method(_osc_server, "/error", "ss", OscClient::_error_handler, this);
	
	_pingack = false;
	_waiting = 0;
	_failed = false;
	_engine_pid = 0;

	_updatetimer = new OscUpdateTimer(this);

	init_traffic_thread();
}


OscClient::~OscClient()
{
	terminate_traffic_thread();

	disconnect(false);

	if (_osc_server) {
		lo_server_free (_osc_server);
	}
	
	delete _updatetimer;
}

bool
OscClient::init_traffic_thread()
{
	// thread for watching for new osc traffic

	_traffic_done = false;
	
	if (pipe (_traffic_request_pipe)) {
		cerr << "Cannot create midi request signal pipe" <<  strerror (errno) << endl;
		return false;
	}

	if (fcntl (_traffic_request_pipe[0], F_SETFL, O_NONBLOCK)) {
		cerr << "UI: cannot set O_NONBLOCK on signal read pipe " << strerror (errno) << endl;
		return false;
	}

	if (fcntl (_traffic_request_pipe[1], F_SETFL, O_NONBLOCK)) {
		cerr << "UI: cannot set O_NONBLOCK on signal write pipe " << strerror (errno) << endl;
		return false;
	}
	
	pthread_create (&_osc_traffic_thread, NULL, &OscClient::_osc_traffic, this);
	if (!_osc_traffic_thread) {
		return false;
	}

	return true;
}

void
OscClient::terminate_traffic_thread ()
{
	void* status;
	char c = 0;

	_traffic_done = true;

	if (write (_traffic_request_pipe[1], &c, 1) != 1) {
		cerr << "cannot send signal to traffic thread! " <<  strerror (errno) << endl;
	}

	pthread_join (_osc_traffic_thread, &status);

	close(_traffic_request_pipe[0]);
	close(_traffic_request_pipe[1]);
}

void
OscClient::setup_param_map()
{
}


bool
OscClient::connect()
{

	if (_osc_addr) {
		lo_address_free (_osc_addr);
	}

	if (_host.empty()) {
		_osc_addr = lo_address_new(NULL, (const char *) wxString::Format(wxT("%ld"), _port).ToAscii());
	} 
	else { 
		_osc_addr = lo_address_new((const char *) _host.ToAscii(), (const char *) wxString::Format(wxT("%ld"), _port).ToAscii());
	}
	//cerr << "osc errstr: " << lo_address_errstr(_osc_addr) << endl;

	_pingack = false;

	_waiting = 0;
	lo_send(_osc_addr, "/ping", "ss", "/pingack", _our_url.c_str());
	//cerr << "sending ping to host: " << _host << " port: " << _port <<  endl;
	_updatetimer->Start(700, true);
	
	/*
	if (!_spawn_config.force_spawn) {
		// send off a ping.  set a timer, if we don't have a response, we'll start our own locally
		_waiting = 0;
		lo_send(_osc_addr, "/ping", "ss", _our_url.c_str(), "/pingack");
		//cerr << "sending ping" << endl;
		_updatetimer->Start(700, true);
	}
	// spawn now
	else if (!_spawn_config.never_spawn && spawn_looper()) {
		//cerr << "immediate spawn" << endl;
		_waiting = 1;
		_updatetimer->Start(100, true);
	}
	else {
		return false;
	}
	*/
	return true;
}

bool
OscClient::disconnect (bool killit)
{

	if (_osc_addr) {
		lo_send(_osc_addr, "/unregister", "ss",  "/param", _our_url.c_str());
		
		if (killit) {
			send_quit();
		}
		
		lo_address_free (_osc_addr);
		_osc_addr = 0;
	}
	_osc_url = "";
	_we_spawned = false;
	
	Disconnected(); // emit
	
	return true;
}


void *
OscClient::_osc_traffic (void *arg)
{
	OscClient * lc = static_cast<OscClient *>(arg);
	lc->osc_traffic();
	return 0;
}

void
OscClient::osc_traffic()
{
	// our only job here in this thread is
	// to send updates when new data comes in on the osc port
	int oscfd = lo_server_get_socket_fd(_osc_server);
	struct pollfd pfd[2];
	int timeout = -1; 
	int nfds = 2;
	struct timespec nsleep = { 0, 20000000 };

	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	pfd[0].fd = _traffic_request_pipe[0];
	pfd[0].events = POLLIN|POLLHUP|POLLERR;
	pfd[1].fd = oscfd;
	pfd[1].events = POLLIN|POLLHUP|POLLERR;
	
	while (!_traffic_done)
	{
		pfd[0].fd = _traffic_request_pipe[0];
		pfd[0].events = POLLIN|POLLHUP|POLLERR;
		pfd[1].fd = oscfd;
		pfd[1].events = POLLIN|POLLHUP|POLLPRI|POLLERR;

		pthread_testcancel();
		
		if (poll (pfd, nfds, timeout) < 0) {
			if (errno == EINTR) {
				/* gdb at work, perhaps */
				// goto again;
			}
			
			cerr << "OSC thread poll failed: " <<  strerror (errno) << endl;
			
			continue;
		}

		if (_traffic_done) {
			break;
		}

		if (nfds > 1 && (pfd[1].revents & POLLIN)) 
		{
			
			// emit signal
			// cerr << "got new data" << endl;
			NewDataReady(); // emit
		}

		// sleep for a bit to give someone the chance to read it
		nanosleep (&nsleep, NULL);

	}
}

bool
OscClient::is_engine_local()
{
	// compare host parts of osc_url and our_url
	bool ret = false;
	char * remhost = lo_url_get_hostname(_osc_url.c_str());
	char * lochost = lo_url_get_hostname(_our_url.c_str());

	// cerr << "osc: " << _osc_url << "  remhost: " << remhost << "  lochost: " << lochost << endl;
	
	if (remhost && lochost && (strncmp(remhost, lochost, 30) == 0)) {
		ret = true;
	}

	free(remhost);
	free(lochost);

	return ret;
}


void
OscClient::pingtimer_expired()
{ 
	update_values();


	// check state of pingack

 	if (_pingack) {
	 	// cerr << "got ping response" << endl;
	} 
	else if (_waiting > 0)
	{
		// NOT USED
	}
	else
	{

		ConnectFailed("Initial connection failed."); // emit
	}

	/*
	 	if (_waiting > 80) {
			if (_lastchance) {
				// give up
				cerr << "slgui: gave up on spawned engine" << endl;
				if (_osc_addr) {
					lo_address_free(_osc_addr);
				}
				_osc_addr = 0;
				_failed = true;
				ConnectFailed("No response from SooperLooper engine process.\nPlease make sure the JACK server is running.\nAlso check that the system's hostname resolves properly."); // emit
			}
			else {
				// last ditch effort with 127.0.0.1 all around
				_osc_addr = lo_address_new("127.0.0.1", (const char *) wxString::Format(wxT("%ld"), _spawn_config.port).ToAscii());

				char tmpbuf[100];
				int sport = lo_server_get_port (_osc_server);
				snprintf (tmpbuf, sizeof(tmpbuf), "osc.udp://127.0.0.1:%d/", sport);
				_our_url = tmpbuf;

				//cerr << "last chance effort: with oururl " << _our_url << endl;
				
				// send off a ping.  
				_pingack = false;
				_waiting = 60;
				lo_send(_osc_addr, "/ping", "ss", _our_url.c_str(), "/pingack");
				_lastchance = 1;
				_updatetimer->Start(100, true);
			}
		}
		else {
			// cerr << "waiting" << endl;
			_waiting++;
			// lo_send(_osc_addr, "/ping", "ss", _our_url.c_str(), "/pingack");
			_updatetimer->Start(100, true);
		}
	}
	else
	{
		// lets try to spawn our own
		if (!_spawn_config.never_spawn && spawn_looper()) {
			cerr << "slgui: spawned new engine" << endl;
			_updatetimer->Start(100, true);
			_waiting = 1;
		}
		else {
			//cerr << "execute failed" << endl;
			if (_osc_addr) {
				lo_address_free(_osc_addr);
			}
			_osc_addr = 0;
			_failed = true;
			if (!_spawn_config.never_spawn) {
				ConnectFailed("Execution of SooperLooper engine process failed."); // emit
			}
		}
			
	}

*/
}


int
OscClient::_pingack_handler(const char *path, const char *types, lo_arg **argv, int argc,
			      void *data, void *user_data)
{
	OscClient * lc = static_cast<OscClient*> (user_data);
	return lc->pingack_handler (path, types, argv, argc, data);
}

int
OscClient::pingack_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data)
{
	// s:hosturl  s:version  i:loopcount

	string hosturl(&argv[0]->s);
	string version(&argv[1]->s);
	char tmpbuf[100];
	long tport = _port;
	
	cerr << "ac3jackgui: remote engine is at " << hosturl << " version=" << version <<  endl;

	
	char * remport = lo_url_get_port(hosturl.c_str());
	wxString tmpstr = wxString::FromAscii(remport);
	tmpstr.ToLong(&tport);
	free(remport);
	_port = tport;
	
	if (_lastchance) {
		// force a 127.0.0.1 hostname if this was on our last chance
		//cerr << "forced a 127" << endl;
		snprintf(tmpbuf, sizeof(tmpbuf), "osc.udp://127.0.0.1:%ld/", _port);
		_osc_url = tmpbuf;
	}
	else {
		_osc_url = hosturl;
	}
	
	char * remhost = lo_url_get_hostname(_osc_url.c_str());
	_host = wxString::FromAscii(remhost);
	free(remhost);



	if (_osc_addr) {
		lo_address_free(_osc_addr);
	}
	_osc_addr = lo_address_new_from_url (hosturl.c_str());

	if (!_pingack) {
		// register future configs with it once
		lo_send(_osc_addr, "/register", "ss",  "/param", _our_url.c_str());

		lo_send(_osc_addr, "/params_all", "ss",  "/param", _our_url.c_str());
		
		_pingack = true;
	}

	return 0;
}

int
OscClient::_alive_handler(const char *path, const char *types, lo_arg **argv, int argc,
			      void *data, void *user_data)
{
	OscClient * lc = static_cast<OscClient*> (user_data);
	return lc->alive_handler (path, types, argv, argc, data);
}

int
OscClient::alive_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data)
{
	// s:hosturl  s:version  i:loopcount

	IsAlive(); // emit
	return 0;
}

int
OscClient::_error_handler(const char *path, const char *types, lo_arg **argv, int argc,
			      void *data, void *user_data)
{
	OscClient * lc = static_cast<OscClient*> (user_data);
	return lc->error_handler (path, types, argv, argc, data);
}

int
OscClient::error_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data)
{
	// s:errstr
	string errsrc(&argv[0]->s);
	string errstr(&argv[1]->s);

	ErrorReceived(errstr); // emit
	return 0;
}



int
OscClient::_param_handler(const char *path, const char *types, lo_arg **argv, int argc,
			      void *data, void *user_data)
{
	OscClient * lc = static_cast<OscClient*> (user_data);
	return lc->param_handler (path, types, argv, argc, data);
}


int
OscClient::param_handler(const char *path, const char *types, lo_arg **argv, int argc, void *data)
{
	// 1st arg is param string, 2nd is value (can be many types)
	if (argc < 2 || types[0] != 's') {
		cerr << "invalid param message" << endl;
		return -1;
	}
	
	wxString ctrl = wxString::FromAscii(&argv[0]->s);
	float fval  = argv[1]->f;
	double dval  = argv[1]->d;
	int ival  = argv[1]->i;
	int64_t i64val  = argv[1]->h;
	wxString sval;

	switch(types[1]) {
	case 'i':
		if ( _params_int_map.find(ctrl) == _params_int_map.end()
		     || _params_int_map[ctrl] != ival) {
			_params_updated[ctrl] = true;
		}
		_params_int_map [ctrl] = ival;
		break;
	case 'h':
		if ( _params_int64_map.find(ctrl) == _params_int64_map.end()
		     || _params_int64_map[ctrl] != i64val) {
			_params_updated[ctrl] = true;
		}
		_params_int64_map [ctrl] = i64val;
		break;
	case 'f':
		if ( _params_float_map.find(ctrl) == _params_float_map.end()
		     || _params_float_map[ctrl] != fval) {
			_params_updated[ctrl] = true;
		}
		_params_float_map [ctrl] = fval;
		break;
	case 'd':
		if ( _params_double_map.find(ctrl) == _params_double_map.end()
		     || _params_double_map[ctrl] != dval) {
			_params_updated[ctrl] = true;
		}
		_params_double_map [ctrl] = dval;
		break;
	case 's':
		sval = wxString::FromAscii(&argv[1]->s);

		if ( _params_string_map.find(ctrl) == _params_string_map.end()
		     || _params_string_map[ctrl] != sval) {
			_params_updated[ctrl] = true;
		}
		_params_string_map [ctrl] = sval;
		break;

	default:
		cerr << "unkonwn type: " << types[1] << endl;
		break;
	}
	

	
	// cerr << "param handler got " << (const char *) ctrl.ToAscii()  << " tpyes: " << types << endl;
	
	return 0;
}


void
OscClient::request_param_value (wxString ctrl)
{
	if (!_osc_addr) return;
	char buf[20];

	snprintf(buf, sizeof(buf), "/param/%s", (const char *) ctrl.ToAscii());
	
	lo_send(_osc_addr, buf, "ss", _our_url.c_str(), "/param");
}


void
OscClient::request_values()
{
	if (!_osc_addr) return;
	char buf[20];

/*
	snprintf(buf, sizeof(buf), "/sl/%d/get", index);
	
	// send request for updates
	lo_send(_osc_addr, buf, "sss", "state", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "loop_pos", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "loop_len", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "cycle_len", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "free_time", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "total_time", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "waiting", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "next_state", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "rate_output", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "channel_count", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "in_peak_meter", _our_url.c_str(), "/ctrl");
	lo_send(_osc_addr, buf, "sss", "out_peak_meter", _our_url.c_str(), "/ctrl");
*/
}




bool
OscClient::post_param_change (wxString ctrl, int val)
{
	if (!_osc_addr) return false;
	char buf[50];

	// go ahead and update our local copy
	_params_int_map[ctrl] = val;
	_params_updated[ctrl] = true;

	snprintf(buf, sizeof(buf), "/param/%s", (const char *) ctrl.ToAscii());

	if (lo_send(_osc_addr, buf, "i", val) == -1) {
		return false;
	}

	return true;
}

bool
OscClient::post_param_change (wxString ctrl, int64_t val)
{
	if (!_osc_addr) return false;
	char buf[50];

	// go ahead and update our local copy
	_params_int64_map[ctrl] = val;
	_params_updated[ctrl] = true;

	snprintf(buf, sizeof(buf), "/param/%s", (const char *) ctrl.ToAscii());

	if (lo_send(_osc_addr, buf, "h", val) == -1) {
		return false;
	}

	return true;
}

bool
OscClient::post_param_change (wxString ctrl, float val)
{
	if (!_osc_addr) return false;
	char buf[50];

	// go ahead and update our local copy
	_params_float_map[ctrl] = val;
	_params_updated[ctrl] = true;

	snprintf(buf, sizeof(buf), "/param/%s", (const char *) ctrl.ToAscii());

	if (lo_send(_osc_addr, buf, "f", val) == -1) {
		return false;
	}

	return true;
}

bool
OscClient::post_param_change (wxString ctrl, double val)
{
	if (!_osc_addr) return false;
	char buf[50];

	// go ahead and update our local copy
	_params_double_map[ctrl] = val;
	_params_updated[ctrl] = true;

	snprintf(buf, sizeof(buf), "/param/%s", (const char *) ctrl.ToAscii());

	if (lo_send(_osc_addr, buf, "d", val) == -1) {
		return false;
	}

	return true;
}

bool
OscClient::post_param_change (wxString ctrl, const wxString & val)
{
	if (!_osc_addr) return false;
	char buf[50];

	// go ahead and update our local copy
	_params_string_map[ctrl] = val;
	_params_updated[ctrl] = true;

	snprintf(buf, sizeof(buf), "/param/%s", (const char *) ctrl.ToAscii());

	if (lo_send(_osc_addr, buf, "s", (const char *) val.ToAscii()) == -1) {
		return false;
	}

	return true;
}


bool
OscClient::post_load_session(wxString fname)
{
	if (!_osc_addr) return false;
	char buf[30];

	snprintf(buf, sizeof(buf), "/load_session");

	// send request for updates
	if (lo_send(_osc_addr, buf, "sss", (const char *) fname.ToAscii(), _our_url.c_str(), "/error") == -1) {
		return false;
	}

	return true;
}

bool
OscClient::post_save_session(wxString fname)
{
	if (!_osc_addr) return false;
	char buf[30];

	snprintf(buf, sizeof(buf), "/save_session");

	// send request for updates
	if (lo_send(_osc_addr, buf, "sss", (const char *) fname.ToAscii(), _our_url.c_str(), "/error") == -1) {
		return false;
	}

	return true;
}


void
OscClient::send_quit()
{
	if (!_osc_addr) return;
	lo_send(_osc_addr, "/quit", NULL);
}

void
OscClient::send_alive_ping()
{
	if (!_osc_addr) return;
	lo_send(_osc_addr, "/ping", "ss",  "/alive_resp", _our_url.c_str());
}

void
OscClient::update_values()
{
	if (!_osc_server) return;
	// recv commands nonblocking, until none left
	//cerr << "updae values" << endl;
	while (lo_server_recv_noblock (_osc_server, 0) > 0)
	{
		// do nothing
	}
	//cerr << "finished updating values" << endl;
}


bool
OscClient::is_updated (wxString ctrl)
{

	UpdatedCtrlMap::iterator iter = _params_updated.find (ctrl);
	
	if (iter != _params_updated.end()) {
		return (*iter).second;
	}

	return false;
}
	
bool
OscClient::get_value (wxString ctrl, float & retval)
{
	bool ret = false;

	FloatControlMap::iterator iter = _params_float_map.find (ctrl);

	if (iter != _params_float_map.end()) {
		retval = (*iter).second;
		// set updated to false
		_params_updated[ctrl] = false;
		ret = true;
	}

	return ret;
}

bool
OscClient::get_value (wxString ctrl, double & retval)
{
	bool ret = false;

	DoubleControlMap::iterator iter = _params_double_map.find (ctrl);

	if (iter != _params_double_map.end()) {
		retval = (*iter).second;
		// set updated to false
		_params_updated[ctrl] = false;
		ret = true;
	}

	return ret;
}

bool
OscClient::get_value (wxString ctrl, int & retval)
{
	bool ret = false;

	IntControlMap::iterator iter = _params_int_map.find (ctrl);

	if (iter != _params_int_map.end()) {
		retval = (*iter).second;
		// set updated to false
		_params_updated[ctrl] = false;
		ret = true;
	}

	return ret;
}

bool
OscClient::get_value (wxString ctrl, wxString & retval)
{
	bool ret = false;

	StringControlMap::iterator iter = _params_string_map.find (ctrl);

	if (iter != _params_string_map.end()) {
		retval = (*iter).second;
		// set updated to false
		_params_updated[ctrl] = false;
		ret = true;
	}

	return ret;
}

bool
OscClient::get_value (wxString ctrl, int64_t& retval)
{
	bool ret = false;

	Int64ControlMap::iterator iter = _params_int64_map.find (ctrl);

	if (iter != _params_int64_map.end()) {
		retval = (*iter).second;
		// set updated to false
		_params_updated[ctrl] = false;
		ret = true;
	}

	return ret;
}
